from ursina import *
from ursina.ursinamath import sample_gradient
from ursina.prefabs.particle_system import play_particle_system


class SplashScreen(Sprite):
    '''A simple splash screen that shows a static image'''

    def __init__(self, texture='ursina_logo', **kwargs):
        super().__init__(parent=camera.ui, texture=texture, world_z=camera.overlay.z-1, scale=.1, color=color.clear, **kwargs)

    def on_window_ready(self):
        camera.overlay.color = color.black
        self.animate_color(color.white, duration=2, delay=1, curve=curve.out_quint_boomerang)
        invoke(destroy, self, delay=3)


    def input(self, key):
        if key in ('space', 'gamepad a', 'escape', 'left mouse down'):
            destroy(self)


    def on_destroy(self):
        camera.overlay.animate_color(color.clear, duration=.25)


class UrsinaSplashScreen(Entity):
    def __init__(self):
        super().__init__(self, parent=camera.ui, world_z=camera.overlay.z-1, name='UrsinaSplashScreen')
        self.logo = Entity(parent=self, model='ursina_logo_wireframe', original_scale=.1, scale=.1, alpha=0)
        self.logo.model.colors = [hsv(lerp(240, 290, (v[0]+.5 + (v[1]*.25))), .8, .75) for v in self.logo.model.vertices]
        self.logo.model.generate()

        text = 'ursina engine'
        hues = (color.hex('#0c72d8'), color.hex('#a964df'), color.hex('#cc38a5'))
        colors = [sample_gradient(hues, i/len(text)) for i in range(len(text))]
        hsv_tags = [f'hsv({','.join([str(e) for e in colr.hsv])})' for colr in colors]

        text = ''.join([f'<{hsv_tags[i]}>{char}' for i, char in enumerate(text)])
        self.text_entity = Text(parent=self, text=text, origin=(0,0), y=-.175, original_scale=3.5, enabled=False)


    def on_destroy(self):
        camera.overlay.animate_color(color.clear, duration=.25)


    def input(self, key):
        if key in ('space', 'gamepad a', 'escape', 'left mouse down'):  # interrupt
            destroy(self)


    def on_window_ready(self):
        camera.overlay.color = color.black
        self.animations.append(invoke(self.play_animation, delay=.2))


    def play_animation(self):
        play_particle_system('ursina_splash_flourish', parent=self.logo, y=-1, scale=2)
        self.text_entity.enabled = False
        self.text_entity.scale = self.text_entity.original_scale
        self.logo.scale_y = 0
        self.logo.fade_in(duration=.05)
        self.logo.animate_scale_y(self.logo.original_scale, duration=3, curve=curve.out_elastic)

        @after(.2, entity=self, unscaled=False)
        def _show_text():
            self.text_entity.enabled = True
            self.text_entity.appear(.025)

        # from ursina.prefabs.ursfx import ursfx
        # from ursina.audio import pitch_note
        # ursfx([(0.0, 1.0), (0.12, 0.5), (0.25, 0.5), (0.46, 0.5), (1.0, 0.0)], volume=0.75, wave='sine', pitch=20, speed=2.7)
        # self.animations.append(invoke(ursfx, [(0.0, 0.0), (0.21, 0.54), (0.45, 0.76), (0.6, 0.75), (1.0, 0.0)], volume=0.53, wave='noise', pitch_change=-12, speed=2.0, delay=1))

        @after(1, entity=self, unscaled=False)
        def _close():
            self.logo.animate_rotation_y(-360, duration=.5, delay=.1, curve=curve.in_out_sine)
            self.logo.animate_scale_y(0, duration=.25, delay=.1, curve=curve.in_elastic)
            self.logo.fade_out(duration=.5)
            self.text_entity.animate_scale_y(0, delay=.4, duration=.2, curve=curve.in_back)
            self.text_entity.animate_scale_x(self.text_entity.original_scale*1.5, delay=.55, duration=.1, curve=curve.out_expo)
            invoke(self.on_destroy, delay=.65)



if __name__ == '__main__':
    app = Ursina(size=Vec2(1920,1080), )

    window.color = color.black
    ursina_splash_screen = UrsinaSplashScreen()

    # application.calculate_dt = False
    # time.dt =  1/60
    # from ursina.prefabs.video_recorder import VideoRecorder
    # vr = VideoRecorder(max_duration=3, fps=60, name='splash_screen')
    # def on_window_ready():
    #     vr.start_recording()


    # EditorCamera()



    # add a custom splash screen after the first one
    # ursina_splash.on_destroy = Func(SplashScreen, 'shore')
    app.run()



import sys
sys.path.append("..")
from ursina import *

class MenuToggler():

    def __init__(self):
        self.target = None


    def on_click(self):
        # close menus before opening a new one
        # if self.entity.parent == scene.editor.load_menu_parent:
        #     for button in scene.editor.load_menu_parent.children:
        #         if button.menu_toggler:
        #             button.menu_toggler.target.enabled = False

        if self.target:
            self.target.enabled = not self.target.enabled
        else:
            for c in self.entity.children:
                c.enabled = not c.enabled



class HitInfo:
    __slots__ = ['hit', 'entity', 'point', 'world_point', 'distance', 'normal', 'world_normal', 'hits', 'entities']

    def __init__(self, **kwargs):
        self.hit:bool = None
        self.entity:Entity = None
        self.point:Vec3 = None
        self.world_point:Vec3 = None
        self.distance:float = 9999
        self.normal:Vec3 = None
        self.world_normal:Vec3 = None
        self.hits = []
        self.entities = []

        for key, value in kwargs.items():
            setattr(self, key, value)


    def __bool__(self):
        return self.hit



import ast
import importlib
import pkgutil
import textwrap
import tokenize
import tomllib
from io import StringIO
from pathlib import Path

from ursina.string_utilities import snake_to_camel


class ClassInfo:
    def __init__(self):
        self.name = None
        self.module = None
        self.base_classes = []
        self.class_variables = []
        self.init_args = []
        self.instance_variables = []
        self.methods = []
        self.properties = []
        self.examples = []
        self.line_number = None

class ModuleInfo:
    def __init__(self):
        self.name = ''
        self.path = ''
        self.imports = []
        self.variables = []
        self.functions = []
        self.classes = []
        self.tests = []
        self.examples = []
        self.is_singleton = False   # this is ursina specific. if a module uses a class (same name as module, but pascal cased), and a 'instance' variable, we should render the class info instead of the module info. for example, window, camera and mouse use this technique to have properties even though they look like modules to the user.

class VarInfo:
    def __init__(self, name='', type_hint=None, default_value=..., comment='', line_number:int=None):
        self.name = name
        self.type_hint = type_hint
        self.default_value = default_value
        self.examples = []
        self.comment = comment
        self.line_number = line_number


class FunctionInfo:
    def __init__(self, name='', type_hint=None, input_args:[VarInfo]=..., comment='', decorators=None, module=None, from_class=None, line_number:int=None):
        self.module = module
        self.from_class = from_class
        self.name = name
        self.type_hint = type_hint
        self.input_args = input_args
        self.comment = comment
        self.decorators = decorators or []
        self.examples = []
        self.line_number = line_number

class Example:
    def __init__(self, name='', content=''):
        self.name = name
        self.content = content


def extract_var_info(node):
    # Regular assignment
    if isinstance(node, ast.Assign):
        for target in node.targets:
            # Module/class level: x = ...
            if isinstance(target, ast.Name):
                return VarInfo(
                    name=target.id,
                    default_value=ast.unparse(node.value) if node.value else None,
                    line_number=target.lineno,
                )
            # # else:
            # #     print(isinstance(target, ast.Attribute), isinstance(target.value, ast.Name), target.value.id)
            # # Instance variable: self.x = ...
            # if isinstance(target, ast.Attribute) and isinstance(target.value, ast.Name) and target.value.id == 'self':
            #     return VarInfo(
            #         name=target.value.id + target.attr,
            #         default_value=ast.unparse(node.value) if node.value else None
            #     )

    # Annotated assignment (with type hint)
    elif isinstance(node, ast.AnnAssign):
        target = node.target
        # Module/class level: x: int = ...
        if isinstance(target, ast.Name):
            return VarInfo(
                name=target.id,
                type_hint=ast.unparse(node.annotation) if node.annotation else None,
                default_value=ast.unparse(node.value) if node.value else None,
                line_number=target.lineno,
            )

        # else:
        #     print(isinstance(target, ast.Attribute), isinstance(target.value, ast.Name), target.value.id)
        # Instance variable: self.x: int = ...
        if isinstance(target, ast.Attribute) and isinstance(target.value, ast.Name) and target.value.id == 'self':
            print('self.', target.attr)
            return VarInfo(
                name=target.attr,
                type_hint=ast.unparse(node.annotation) if node.annotation else None,
                default_value=ast.unparse(node.value) if node.value else None,
                line_number=target.lineno,
            )

    return None


def extract_function_args(func_node):
    args = func_node.args.args
    defaults = func_node.args.defaults

    num_args = len(args)
    num_defaults = len(defaults)
    default_offset = num_args - num_defaults

    result = []

    for i, arg in enumerate(args):
        var_info = VarInfo()
        var_info.name = arg.arg
        var_info.type_hint = ast.unparse(arg.annotation) if arg.annotation else None

        # Match defaults from the right
        if i >= default_offset:
            default_expr = defaults[i - default_offset]
            var_info.default_value = ast.unparse(default_expr)
        else:
            var_info.default_value = None

        result.append(var_info)

    return result




def extract_inline_comments(source_code):
    comments = {}
    tokens = tokenize.generate_tokens(StringIO(source_code).readline)
    for tok_type, tok_string, (lineno, _), _, _ in tokens:
        if tok_type == tokenize.COMMENT:
            comments[lineno] = tok_string.lstrip('#').strip()
    return comments

def extract_example(str, name=None):    # use name to highlight the relevant class
    key = '''if __name__ == '__main__':'''
    lines = []
    example_started = False
    for l in str.split('\n'):
        if example_started:
            lines.append(l)

        if l == key:
            example_started = True

    example = '\n'.join(lines)
    example = textwrap.dedent(example)
    # example = example.split('# test\n')[0]
    # ignore = ('app = Ursina()', 'app.run()', 'from ursina import *')
    ignore = ()
    # if 'class Ursina' in str:   # don't ignore in main.py
    #     ignore = ()

    lines = [e for e in example.split('\n') if not e in ignore and not e.strip().startswith('#')]
    text = '\n'.join(lines)
    return text


def is_singleton(source_code, module_name):
    for l in source_code.split('\n'):
        if l.startswith(f'instance = {snake_to_camel(module_name).title()}('):
            return True
    return False

def get_package_version(package_path:Path):
    with (package_path / 'pyproject.toml').open('rb') as f:
        data = tomllib.load(f)

    version = data['project']['version']
    return version


def analyze_module(source_code: str, path:Path=None) -> ModuleInfo:
    tree = ast.parse(source_code)
    inline_comments = extract_inline_comments(source_code)

    module_info = ModuleInfo()
    module_info.path = path
    module_info.examples.append(Example(name='', content=extract_example(source_code)))
    module_info.is_singleton = is_singleton(source_code, path.stem) if path is not None else False

    for node in tree.body:
         # module level variables
        if isinstance(node, ast.Assign | ast.AnnAssign):
            var_info = extract_var_info(node)
            if var_info:
                module_info.variables.append(var_info)

        # module level functions
        elif isinstance(node, ast.FunctionDef):
            module_info.functions.append(FunctionInfo(
                name = node.name,
                module = module_info,
                type_hint = ast.unparse(node.returns) if node.returns else None,
                input_args = extract_function_args(node),
                decorators = [ast.unparse(d) for d in node.decorator_list],
                comment = inline_comments.get(node.lineno),
                line_number=node.lineno,
                )
            )

        # classes
        if isinstance(node, ast.ClassDef):
            class_info = ClassInfo()
            module_info.classes.append(class_info)
            class_info.name = node.name
            class_info.base_classes = [ast.unparse(base) for base in node.bases] if node.bases else []


            for statement in node.body:
                # class level variables
                if isinstance(statement, ast.Assign):
                    for target in statement.targets:
                        if isinstance(target, ast.Name):
                            name = target.id
                            default_value = ast.unparse(statement.value) if statement.value else None
                            class_info.class_variables.append(VarInfo(
                                name=name,
                                default_value=default_value,
                                comment=inline_comments.get(target.lineno),
                                line_number=target.lineno,
                                ))

                # class level variables with type hint
                elif isinstance(statement, ast.AnnAssign) and isinstance(statement.target, ast.Name):
                    class_info.class_variables.append(extract_var_info(statement))

                # methods
                elif isinstance(statement, ast.FunctionDef):
                    method_name = statement.name
                    class_info.methods.append(FunctionInfo(
                        module = module_info,
                        from_class = class_info,
                        name = statement.name,
                        # type_hint = ast.unparse(node.returns) if node.returns else None,
                        input_args = [e for e in extract_function_args(statement) if e.name != 'self'],
                        decorators = [ast.unparse(d) for d in statement.decorator_list],
                        comment = inline_comments.get(statement.lineno),
                        line_number = statement.lineno,
                        )
                    )

                    if method_name == '__init__':
                        # Find instance variables: self.xxx = ...
                        for init_statement in ast.walk(statement):
                            if isinstance(init_statement, ast.Assign):
                                for target in init_statement.targets:
                                    if isinstance(target, ast.Attribute) and isinstance(target.value, ast.Name) and target.value.id == 'self':
                                        class_info.instance_variables.append(VarInfo(
                                            name=target.attr,
                                            default_value=ast.unparse(init_statement.value) if init_statement.value else None,
                                            comment=inline_comments.get(target.lineno),
                                            line_number=statement.lineno,
                                        ))
    return module_info



def get_all_modules_in_package(package):
    modules = []
    for _importer, modname, _ispkg in pkgutil.walk_packages(package.__path__, package.__name__ + "."):
        try:
            # print('Importing:', modname)
            module = importlib.import_module(modname)
            modules.append(module)
        except Exception as e:
            print(f"Failed to import {modname}: {e}")
    return modules


if __name__ == '__main__':
    from ursina import application
    # with open(application.package_folder / 'application.py', 'r') as f:
    path = application.package_folder / 'entity.py'
    with path.open('r') as f:
        module_info = analyze_module(f.read())

    print('module name:', module_info.name)
    print('--- module variables ---')
    for e in module_info.variables:
        # print('module level var:', e)
        if e.name.startswith('_'):
            continue
        type_hint = f': {e.type_hint}' if e.type_hint is not None else ''
        print(f'    {e.name} {type_hint}= {e.default_value}')

    print('--- module functions ---')
    for function_info in module_info.functions:
        print(function_info)

    for class_info in module_info.classes:
        print('--- class ---')
        print(f'class: {class_info.name}({', '.join(class_info.base_classes)})')
        print('--- class_variables ---')
        for e in class_info.class_variables:
            if e.name.startswith('_'):
                continue
            # print(f'    {class_info.name}.{e.name} = ')
            type_hint = f': {e.type_hint}' if e.type_hint is not None else ''
            print(f'    {e.name} {type_hint}= {e.default_value}')

        print('--- init ---')
        print(f'{class_info.name}(')
        for e in class_info.init_args:
            if e.name.startswith('_'):
                continue
            type_hint = f': {e.type_hint}' if e.type_hint is not None else ''
            print(f'    {e.name} {type_hint}= {e.default_value}')
        print('    )')

        print('--- properties ---')
        for e in class_info.properties:
            print('   ', e)
        # print('args:', class_info.init_args)




from panda3d.core import Shader


empty_shader = Shader.make(Shader.SL_GLSL,
vertex='''
#version 130
uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 texcoord;

void main() {
  gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
  texcoord = p3d_MultiTexCoord0;
}
''',

fragment='''
#version 130

uniform sampler2D p3d_Texture0;
in vec2 texcoord;
out vec4 fragColor;


void main() {
    // gl_FragColor.rgb = vec3(0,1,1);
    vec4 color = texture(p3d_Texture0, texcoord);
    fragColor = color.rgba;
}


''', geometry='')


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='cube', color=color.orange)
    camera.shader = empty_shader
    EditorCamera()

    app.run()



from ursina import *

def length(v):
    return sqrt(v.x*v.x + v.y*v.y)

def signed_distance_to_circle(point, center, radius):
    return length(center-point) - radius


app = Ursina()

random.seed(1)
camera.orthographic = True
camera.fov = 1
window.color = color.black
bg = Entity(model='quad', scale=(window.aspect_ratio, 1) * 1000, collider='box', color=color._32, visible=False)
origin = Entity()
origin_renderer = Entity(parent=origin, model='sphere', color=color.yellow, scale=.025)
line = Entity(parent=origin, model='cube', origin_z=-.5, scale=(.0025, .0025, 999), color=color.gray)
current_point = Entity(parent=origin, model='sphere', color=color.white33)
world_space_cursor = Entity(model='circle', color=color.pink, scale=.0125, always_on_top=True)
mouse.visible = False

objects = list()
for i in range(8):
    objects.append(Entity(model='circle', scale=random.uniform(.05, .3), x=random.random()-.5, y=random.random()-.5, color=color.gray))


sphere_renderers = [Entity(model='circle', color=color.color(180,1,1,.25), enabled=False) for i in range(100)]


def update():
    world_space_cursor.position = mouse.world_point
    origin.look_at(world_space_cursor)
    origin.x += (held_keys['d'] - held_keys['a']) * time.dt * .25
    origin.y += (held_keys['w'] - held_keys['s']) * time.dt * .25



def input(key):
    if key == 'left mouse down':
        # reset
        current_point.position = (0, 0, 0)
        for sphere in sphere_renderers:
            sphere.enabled = False

        # sphere cast
        for i, sphere in enumerate(sphere_renderers):
            distances = [signed_distance_to_circle(current_point.world_position, e.world_position, e.scale_x/2) for e in objects]
            new_sphere_radius = min(distances)
            if new_sphere_radius > 999: # hit nothing
                break

            if new_sphere_radius < .001:
                print(f'collided after {i} steps')
                break

            sphere.enabled = True
            sphere.scale = new_sphere_radius * 2
            sphere.world_position = current_point.world_position
            current_point.z += new_sphere_radius


app.run()



from ursina import *

app = Ursina()

ground = Entity(model='plane', scale=10, color=color.gray, collider='box')
cube = Entity(model='cube', color=color.azure, position=(0,0.5,0))
directional_light = DirectionalLight(shadows=True, color=color.rgb32(255, 221, 200))
# directional_light.look_at(Vec3(1, -1, -1))
# directional_light.color = color.white

EditorCamera()

app.run()


from ursina import *

class RebindMenu(Entity):
    def __init__(self, **kwargs):
        super().__init__(parent=camera.ui, ignore_paused=True)

        self.bg = Panel(parent=self)

        for key, value in kwargs.items():
            setattr(self, key ,value)


        self.fields = []

        for i, (name, value) in enumerate({'jump':'space', 'left':'d', 'right':'a'}.items()):
            self.fields.append(Button(parent=self, scale=.05, y=.45-(i*.06)))



if __name__ == '__main__':
    app = Ursina()
    RebindMenu()
    app.run()



from ursina import *

app = Ursina()
voxels = list()

blank = '00000\n' * 5
blank = [[bool(i=='1') for i in e] for e in blank.strip().split('\n')]
print(blank)

slice = '''
00000
01110
01110
01110
00000
'''
slice = [[bool(i=='1') for i in e] for e in slice.strip().split('\n')]
# TODO: add empty space around

plane = [(0,0,0), (1,0,0), (1,1,0), (0,0,0), (0,1,0), (1,1,0)]
plane = [Vec3(e) for e in plane]
plane = [e - Vec3(.5,.5,.5) for e in plane]
right_plane = Entity(model=Mesh(plane), color=color.white33, double_sided=True)
for v in plane:
    Entity(parent=right_plane, position=v, scale=.1, model='cube', color=color.red)


# shapes = {
#     '011111' : get_verts((1,0,0)),     # flat
#     '101111' : get_verts((-1,0,0)),
#     '110111' : get_verts((0,1,0)),
#     '111011' : get_verts((0,-1,0)),
#     # ''          # edge
#     # ''          # outer corner
#     # ''          # inner corner
# }
verts = list()
mesh = Mesh(vertices=verts)
result = Entity(model=mesh)

def add_flat(direction=(0,0,1)):
    print('adding', direction)
    right_plane.look_at(direction, 'back')
    # [verts.append(c.world_position) for c in right_plane.children]
    for c in right_plane.children:
        verts.append(c.world_position)
        # print(c.world_position)
        # invoke(setattr, c, 'color', color.green, delay=i*.5)
        # invoke(setattr, c, 'color', color.red, delay=(i*.5)+.1)

    mesh.vertices = verts
    mesh.generate()


voxels = (blank, slice, slice, slice, blank)
dir_dict = {'11':0, '01':1, '10':-1}

i = 0
for z in range(len(voxels)):
    for y in range(len(voxels[0])):
        for x in range(len(voxels[0][0])):
            if not voxels[x][y][z]:
                continue

            neighbours = (
                voxels[x+1][y][z],
                voxels[x-1][y][z],
                voxels[x][y+1][z],
                voxels[x][y-1][z],
                voxels[x][y][z+1],
                voxels[x][y][z-1],
                )
            neighbours = ''.join([str(int(e)) for e in neighbours])
            # if neighbours in shapes:
            if neighbours.count('0') == 1:
            # print(len(shapes[neighbours]))
                direction=(dir_dict[neighbours[0:2]], dir_dict[neighbours[2:4]], dir_dict[neighbours[4:6]])
                invoke(add_flat, direction, delay=i*.05)
            # for v in shapes[neighbours]:
            #     verts.append(v)
            #     print(v)
            i += 1




print(verts)
# cube_verts = ((0,0,0), (1,0,0), (0,0,1), (1,0,1), (0,1,0), (1,1,0), (0,1,1), (1,1,1))
# cube_tris =  ((0,2,3,1), (0,1,5,4), (1,3,7,5), (3,2,6,7), (2,0,4,6), (5,7,6,4))
# tris = list()
# for e in cube_tris:
#     tris.append((e[0],e[1],e[2]))
#     tris.append((e[2],e[3],e[0]))

# print(tris)
# Entity(model=Mesh(vertices=verts))
# Entity(model=Mesh(vertices=cube_verts, triangles=tris))
EditorCamera()


# def bevel_point(verts, tris, index, amount=.5):
#     connected_points = list()
#     # for i in range(len(verts)):
#     #     if i == index:
#     #         continue
#     connected_tris = [t for t in tris if index in t]
#     for tri in connected_tris:
#         for vertex_index in tri:
#             if not vertex_index == index:
#                 # print(verts[vertex_index])
#                 pos = lerp(verts[index], verts[vertex_index], amount)
#                 Entity(model='cube', scale=.05, color=color.red, position=pos)
#
#     # printvar(connected_tris)
# bevel_point(cube_verts, tris, 1)

app.run()



'''
clicker game
make a gold counter
make a button
you earn gold for every click
when you have enough gold you can unlock new nodes to automatically generate gold!
'''



from ursina import *

app = Ursina()
window.color = color._20

gold = 0
counter = Text(text='0', y=.25, z=-1, scale=2, origin=(0,0), background=True)
button = Button(text='+', color=color.azure, scale= .125)

def button_click():
    global gold
    gold += 1
    counter.text = str(gold)

button.on_click = button_click



button_2 = Button(cost=10, x=.2, scale=.125, color=color.dark_gray, disabled=True)
button_2.tooltip = Tooltip(f'<gold>Gold Generator\n<default>Earn 1 gold every second.\nCosts {button_2.cost} gold.')

def buy_auto_gold():
    global gold
    if gold >= button_2.cost:
        gold -= button_2.cost
        counter.text = str(gold)
        invoke(auto_generate_gold, 1, 1)

button_2.on_click = buy_auto_gold



def auto_generate_gold(value=1, interval=1):
    global gold
    gold += 1
    counter.text = str(gold)
    button_2.animate_scale(.125 * 1.1, duration=.1)
    button_2.animate_scale(.125, duration=.1, delay=.1)
    invoke(auto_generate_gold, value, delay=interval)


def update():
    global gold
    for b in (button_2, ):
        if gold >= b.cost:
            b.disabled = False
            b.color = color.green
        else:
            b.disabled = True
            b.color = color.gray



app.run()



import bpy
import os
import sys


blend_file_path = bpy.data.filepath
filepath = sys.argv[-1]
mtl_pattern = '*.mtl' if '--export_mtl' in sys.argv else ''

bpy.ops.wm.obj_export(
    filepath=filepath,
    check_existing=True,
    forward_axis='Z',
    up_axis='Y',
    filter_glob=f"*.obj;{mtl_pattern}",
    export_selected_objects=False,
    export_smooth_groups=False,
    export_normals=True,
    export_uv=True,
    export_materials=('--export_mtl' in sys.argv),
    export_colors=True,
    export_triangulated_mesh=True,
    export_curves_as_nurbs=False,
    export_vertex_groups=False,
    global_scale=1,
    path_mode='AUTO'
)



from ursina import *






class TransformGizmo(Draggable):
    def __init__(self, **kwargs):
        super().__init__(
            model='quad',
            scale=1/9/4 * .75,
            step=0,
            origin=(-.5,-.5),
            min_x=-.5*camera.aspect_ratio*.75,
            max_x=.5*camera.aspect_ratio*.75,
            min_y=-.5*.75,
            max_y=.5*.75,
            )

        self.x_axis_ruler = Entity(parent=self, model='quad', scale=(9999,.1), z=.1, color=color.red, alpha=.3)
        self.y_axis_ruler = Entity(parent=self, model='quad', scale=(9999,.1), z=.1, color=color.lime, rotation_z=90, alpha=.3)

        self.right_arrow = Draggable(parent=self, model='quad', origin=(-.5,-.5), color=color.orange, scale_y=.15, scale_x=3, lock_x=True, lock_y=True)
        def horizontal_drag():
            mouse.position = self.position
            self.lock_y = True
            self.dragging = True
        self.right_arrow.drag = horizontal_drag

        self.up_arrow = Draggable(parent=self, model='quad', origin=(-.5,.5), color=color.lime, scale_y=.15, scale_x=3, lock_x=True, lock_y=True, rotation_z=-90)
        def vertical_drag():
            mouse.position = self.position
            self.lock_x = True
            self.dragging = True
        self.up_arrow.drag = vertical_drag

        def drop():
            self.lock_x = False
            self.lock_y = False
            self.dragging = False

        self.right_arrow.drop = drop
        self.up_arrow.drop = drop


        for key, value in kwargs.items():
            setattr(self, key, value)



    def update(self):
        super().update()

        self.x_axis_ruler.enabled = held_keys['x']
        self.y_axis_ruler.enabled = held_keys['y']

        if held_keys['x']:
            self.world_y = self.start_pos[1] # start pos is inherited from Draggable
        if held_keys['y']:
            self.world_x = self.start_pos[0] # start pos is inherited from Draggable



class UIDesigner(Entity):
    def __init__(self, **kwargs):
        super().__init__(parent=camera.ui)
        self.scale *= .75
        self.transform_gizmo = TransformGizmo()
        self.bg = Entity(parent=self, model='quad', scale_x=camera.aspect_ratio, color=window.color, z=100)
        self.grid = Entity(
            parent=self,
            position=window.bottom_left,
            model=Grid(16*8, 9*8),
            scale_x=camera.aspect_ratio,
            z=-100,
            color=color.color(0,0,1,.1),
            enabled=False
            )
        self.snap = False
        self.ui_elements = list()
        self.selection = None
        window.color = color._32

    def input(self, key):
        if key == 'shift':
            self.snap = not self.snap

        if key == 'shift up':
            self.snap = not self.snap

        if key == 's':
            self.snap = not self.snap

        if key == 'b':
            self.add_button()

        if key == 'left mouse down':
            self.transform_gizmo.enabled = mouse.hovered_entity
            self.transform_gizmo.position = self.selection.get_position(relative_to=self)

            if not mouse.hovered_entity and self.selection:
                self.selection.world_parent = self.selection.org_parent
                self.selection = None
                self.transform_gizmo.enabled = False
                return

            if mouse.hovered_entity not in self.ui_elements or mouse.hovered_entity == self.selection:
                return

            self.selection = mouse.hovered_entity

            self.transform_gizmo.z = -100
            self.selection.org_parent = self.selection.parent
            self.selection.org_z = self.selection.z
            self.selection.world_parent = self.transform_gizmo
            print_on_screen('select: ' + self.selection.name)

        if key == 'left mouse up' and self.selection and mouse.hovered_entity == self.transform_gizmo:
            self.selection.world_parent = self.selection.org_parent
            self.selection.z = self.selection.org_z




    @property
    def snap(self):
        return self._snap

    @snap.setter
    def snap(self, value):
        self._snap = value
        if value:
            self.grid.enabled = not self.grid.enabled
            if self.grid.enabled:
                self.transform_gizmo.step = (1/9/8*self.scale_y, 1/9/4*self.scale_y)
        else:
            self.grid.enabled = False
            self.transform_gizmo.step = 0

    def add_button(self):
        self.ui_elements.append(Button(parent=self, scale=(.25,.05), text='Button', ignore=True))




if __name__ == '__main__':
    app = Ursina()
    UIDesigner()
    app.run()



from ursina import *


# window.size = (128,128)
app = Ursina()
#
p = Entity()
# Entity(parent=p, model='plane', color=color.gray, scale=50)
Entity(parent=p, model='procedural_rock_0', color=color.dark_gray, scale=2, rotation_z=-45, y=1)
# Entity(parent=p, model='sphere', color=color.dark_gray, scale=4, rotation_z=-45, y=2)
Entity(model='plane', color=color.yellow.tint(-.3), scale=50)


p.combine()
side_color = color.color(44,.43,.26)
p.model.colorize(
    side_color, side_color,
    side_color.tint(-.2), color.gray,
    side_color, side_color,
    smooth=False)
# # p.model.calculate_normals()
# # print(p.bounds)
# # bounds = [int(e) for e in p.bounds]
# # voxels = list()
# #
# # rounded_verts = [(int(v[0]), int(v[1]), int(v[2])) for v in p.model.vertices]
# # print(rounded_verts)
# # for y in range(bounds[1]):
# #     for z in range(bounds[2]):
# #         for x in range(bounds[0]):
# #             if not (x,y,z) in rounded_verts:
# #                 continue
# #
# #             e = Entity(model='cube', origin=(-.5,-.5,-.5), color=color.random_color(), position=(x,y,z))
# #
# #             voxels.append(e)
# seq = Sequence()
# # target = Entity(model='quad', scale=.1, color=color.lime)
# target = camera
#
# new_colors = list()
#
# for i, v in enumerate(p.model.vertices):
#     seq.append(Func(setattr, target, 'position', Vec3(*v)-(Vec3(*p.model.normals[i])*.05)))
#     seq.append(Func(target.look_at, Vec3(*v)+p.model.normals[i]))
#     # seq.append(Wait(.01))
#     # seq.append(Func(app.graphicsEngine.renderFrame))
#
#
# seq.append(Func(EditorCamera))
#
#
# def input(key):
#     if key == 'b':
#         seq.start()
#
#
# # Sky()
# sun = Entity(model='sphere', color=color.yellow, position=(10,100,30), scale=30)
#
#
# # def capture():
#     # t = base.screenshot()
#     # print(t.get_pixel(0,0))
#
#
#     # from panda3d.core import Shader
#     # from direct.filter.FilterManager import FilterManager
#     # # manager = FilterManager(base.win, base.cam)
#     # manager = FilterManager(base.win, camera.ui_camera)
#     # from panda3d.core import Texture as PandaTexture
#     # tex = PandaTexture()
#     # quad = manager.renderSceneInto(colortex=tex)
#     # # quad.setShader(Shader.load("contrast.sha"))
#     # quad.setShaderInput("tex", tex)
#     # # quad.setShaderInput("contrast", 1.0)
# altBuffer = app.win.makeTextureBuffer("hello", 256, 256)
# app.makeCamera(altBuffer)
# Entity(model='quad', texture=altBuffer.getTexture())
#
# window.center_on_screen()
# # p.color = color.black
# # window.color = color.black
# # invoke(capture, delay=1)
EditorCamera()
app.run()



from direct.filter.FilterManager import FilterManager
from panda3d.core import Camera as PandaCamera
from panda3d.core import LensNode, NodePath, OrthographicLens, PerspectiveLens
from panda3d.core import Texture as PandaTexture

from ursina import application, color, entity
from ursina.entity import Entity
from ursina.scene import instance as scene
from ursina.scripts.property_generator import generate_properties_for_class
from ursina.shader import Shader
from ursina.string_utilities import print_info
from ursina.texture import Texture
from ursina.window import instance as window


@generate_properties_for_class()
class Camera(Entity):

    def __init__(self, **kwargs):
        original_warn_if_ursina_not_instantiated = entity._warn_if_ursina_not_instantiated
        entity._warn_if_ursina_not_instantiated = False
        super().__init__(**kwargs)
        entity._warn_if_ursina_not_instantiated = original_warn_if_ursina_not_instantiated

        self.parent = scene
        self.name = 'camera'
        self.eternal = True

        self._cam = None
        self._render = None

        self._ui_lens_node = None
        self._perspective_lens_node = None
        self._orthographic_lens_node = None
        self._ui_size = 40
        self.ui = Entity(eternal=True, name='ui', scale=(self._ui_size*.5, self._ui_size*.5), add_to_scene_entities=False)
        self.overlay = Entity(parent=self.ui, model='quad', scale=99, color=color.clear, eternal=True, z=-99, add_to_scene_entities=False)
        # self.ready = False
        # self._orthographic = False
        # self._fov = 40   # horizontal fov


    def _set_up(self):
        self.display_region = application.base.camNode.get_display_region(0)
        win = self.display_region.get_window()

        self.perspective_lens = PerspectiveLens()
        self.perspective_lens = application.base.camLens # use panda3d's default for automatic aspect ratio on window resize
        self.lens = self.perspective_lens
        self.perspective_lens.set_aspect_ratio(window.aspect_ratio)
        self._perspective_lens_node = LensNode('_perspective_lens_node', self.perspective_lens)
        self.lens_node = self._perspective_lens_node

        self.orthographic_lens = OrthographicLens()
        self.orthographic_lens.set_film_size(self.fov * window.aspect_ratio, self.fov)
        self._orthographic_lens_node = LensNode('_orthographic_lens_node', self.orthographic_lens)

        application.base.cam.node().set_lens(self.lens)

        self.orthographic = False
        self.fov = 40   # horizontal fov
        self.clip_plane_near = 0.1
        self.clip_plane_far = 10000

        self.ui_display_region = win.make_display_region()
        self.ui_display_region.set_sort(20)

        self.ui_camera = NodePath(PandaCamera('ui_camera'))
        self.ui_lens = OrthographicLens()
        # moved set_film_size() to window module for correct aspect ratio after setting window size
        self.ui_lens.set_near_far(-1000, 1000)
        self.ui_camera.node().set_lens(self.ui_lens)
        self._ui_lens_node = LensNode('_ui_lens_node', self.ui_lens)

        self.ui_render = NodePath('ui_render')
        self.ui_render.set_depth_test(True)
        self.ui_render.set_depth_write(True)
        self.ui_camera.reparent_to(self.ui_render)
        self.ui_display_region.set_camera(self.ui_camera)
        scene.ui_camera = self.ui_camera

        self.ui.parent = self.ui_camera

        # these get created when setting a shader
        self.filter_manager = None
        self.filter_quad = None
        self.render_texture = None
        self.filter_quad = None
        self.depth_texture = None


    def orthographic_getter(self):
        return getattr(self, '_orthographic', False)

    def orthographic_setter(self, value):
        self._orthographic = value
        self.lens_node = (self._perspective_lens_node, self._orthographic_lens_node)[value] # this need to be set for the mouse raycasting
        application.base.cam.node().set_lens((self.perspective_lens, self.orthographic_lens)[value])
        self.fov = self.fov

    def fov_getter(self):
        return getattr(self, '_fov', 0)
    def fov_setter(self, value):
        value = max(1, value)
        self._fov = value
        if not self.orthographic and hasattr(self, 'perspective_lens'):
            self.perspective_lens.set_fov(value)
            # self.perspective_lens.set_fov(value*window.aspect_ratio, value)

        elif self.orthographic and hasattr(self, 'orthographic_lens'):
            self.orthographic_lens.set_film_size(value * self.aspect_ratio, value)

    def clip_plane_near_getter(self):
        return self._clip_plane_near
    def clip_plane_near_setter(self, value):
        self._clip_plane_near = value
        self.lens.set_near(value)

    def clip_plane_far_getter(self):
        return self._clip_plane_far
    def clip_plane_far_setter(self, value):
        self._clip_plane_far = value
        self.lens.set_far(value)


    def aspect_ratio_getter(self):      # get current aspect ratio. can not be set.
        return self.perspective_lens.get_aspect_ratio()


    def shader_setter(self, value):     # for applying post-processing effects.
        if value == Entity.default_shader:
            return
        self._shader = value
        if value is None:
            self.filter_manager.cleanup()
            self.filter_manager = None
            if self.filter_quad:
                self.filter_quad.removeNode()
                # print('removed shader')
            return None

        shader = value

        if isinstance(value, Shader) and not shader.compiled:
            shader.compile()

        if isinstance(value, Shader):
            shader = shader._shader

        if not self.filter_manager:
            self.filter_manager = FilterManager(application.base.win, application.base.cam)
            self.render_texture = PandaTexture()
            self.depth_texture = PandaTexture()
            # self.normals_texture = PandaTexture()
            # from panda3d.core import AuxBitplaneAttrib
            self.filter_quad = self.filter_manager.renderSceneInto(
                colortex=self.render_texture,
                depthtex=self.depth_texture,
                # auxtex=self.normals_texture,
                # auxbits=AuxBitplaneAttrib.ABOAuxNormal
                )
            self.filter_quad.set_shader_input("tex", self.render_texture)
            self.filter_quad.set_shader_input("dtex", self.depth_texture)

            if hasattr(value.default_input, 'clip_plane_near'):
                self.clip_plane_near = value.default_input['clip_plane_near']
            if hasattr(value.default_input, 'clip_plane_far'):
                self.clip_plane_far = value.default_input['clip_plane_far']
            # self.filter_quad.set_shader_input("ntex", self.normals_texture)

        self.filter_quad.setShader(shader)

        if hasattr(value, 'default_input'):
            for key, input_value in value.default_input.items():
                if callable(input_value):
                    input_value = input_value()
                self.set_shader_input(key, input_value)


        print_info('set camera shader to:', shader)



    def set_shader_input(self, name, value):
        if not hasattr(self, 'filter_quad') or self.filter_quad is None:
            return

        if isinstance(value, Texture):
            value = value._texture    # make sure to send the panda3d texture to the shader

        self.filter_quad.setShaderInput(name, value)


instance = Camera()


if __name__ == '__main__':
    from ursina import *
    from ursina import EditorCamera, Entity, Ursina, camera

    # window.borderless = False
    app = Ursina()

    camera.orthographic = True

    e = Entity()
    e.model = 'quad'
    e.color = color.random_color()
    e.position = (-2, 0, 10)

    e = Entity()
    e.model = 'quad'
    e.color = color.random_color()
    e.position = (2, 0, 10)

    e = Entity()
    e.model = 'quad'
    e.color = color.random_color()
    e.position = (0, 0, 40)

    EditorCamera()
    # from ursina import *
    # Button(text='test button')
    from ursina.shaders import camera_grayscale_shader
    camera.shader = camera_grayscale_shader

    # def update():
    #     t = Texture(camera.render_texture)
    #     print(t.pixels)
    app.run()



from ursina import Animator, Audio, Button, Entity, Slider, Text, audio, camera, color, hsv, scene
from ursina.prefabs.button_group import ButtonGroup


class MenuButton(Button):
    def __init__(self, text='', **kwargs):
        super().__init__(text, scale=(.25, .075), highlight_color=color.azure, **kwargs)
        for key, value in kwargs.items():
            setattr(self, key ,value)

# button_size = (.25, .075)
button_spacing = .05

class OptionsMenu(Entity):
    default_values = dict(parent=camera.ui)
    def __init__(self, **kwargs):
        super().__init__(**(__class__.default_values | kwargs))

        self.accessibility_menu = Entity(parent=self)
        self.accessibility_menu.preview_text = Text(parent=self.accessibility_menu, x=.275, y=.25, text='Preview text', origin=(-.5,0))
        for t in [e for e in scene.entities if isinstance(e, Text)]:
            t.original_scale = t.scale

        text_scale_slider = Slider(0, 2, default=1, step=.1, dynamic=True, text='Text Size', parent=self.accessibility_menu, x=-.25)
        def set_text_scale():
            for t in [e for e in scene.entities if isinstance(e, Text) and hasattr(e, 'original_scale')]:
                t.scale = t.original_scale * text_scale_slider.value
        text_scale_slider.on_value_changed = set_text_scale

        for i, e in enumerate(self.accessibility_menu.children):
            e.y = .3 + (-i * button_spacing)


        self.audio_settings_menu = Entity(parent=self)
        def update_volume_for_currently_playing():
            for e in scene.entities:
                if isinstance(e, Audio):
                    e.volume = e.volume

        volume_slider = Slider(0, 1, default=Audio.volume_multiplier, step=.1, text='Master Volume', parent=self.audio_settings_menu, x=-.25, dynamic=True)
        def set_volume(slider=volume_slider):
            Audio.volume_multiplier = slider.value
            update_volume_for_currently_playing()
        volume_slider.on_value_changed = set_volume

        for name, audio_group in audio.audio_groups.items():
            volume_slider = Slider(0, 1, default=audio_group.volume_multiplier, step=.1, text=f'{name.title()} Volume', parent=self.audio_settings_menu, x=-.25, dynamic=True)
            def set_volume(name=name, slider=volume_slider):
                audio.audio_groups[name].volume_multiplier = slider.value
                update_volume_for_currently_playing()
            volume_slider.on_value_changed = set_volume

        for i, e in enumerate(self.audio_settings_menu.children):
            e.y = .3 + (-i * button_spacing)

        self.graphics_menu = Entity(parent=self)
        window_mode_setting = ButtonGroup(('Borderless Fullscreen', 'Windowed'), y=.3, origin=(0,.5), parent=self.graphics_menu, label='Window Mode')
        def _set_window_mode():
            if window_mode_setting.value == window_mode_setting.options[0]:
                window.borderless = True
                window.fullscreen = True
            elif window_mode_setting.value == window_mode_setting.options[1]:
                window.borderless = False
                # window.size = window.windowed_size
        window_mode_setting.on_value_changed = _set_window_mode

        window_index_setting = ButtonGroup([f'{i: ^10}' for i in range(len(window.monitors))], y=.3, origin=(0,.5), parent=self.graphics_menu, label='Monitor')

        for i, e in enumerate((window_mode_setting, window_index_setting)):
            e.y = .3 + (-i * button_spacing*3)


        self.language_menu = Entity(parent=self)
        ButtonGroup(('English', 'Mandarin Chinese (官话)', 'Spanish (español)', 'Japanese (日本語 )', 'German (Deutsch)', 'Norwegian (bokmål)'), y=.3, origin=(0,.5), parent=self.language_menu, max_x=1)

        self.controls_menu = Entity(parent=self)

        self.state_handler = Animator({
            'controls' :        self.controls_menu,
            'graphics' :        self.graphics_menu,
            'audio' :           self.audio_settings_menu,
            'accessibility' :   self.accessibility_menu,
            'language' :        self.language_menu,
            }
        )

        # self.side_menu = ButtonList({key.title(): Func(setattr, self.state_handler, 'state', key) for key, value in self.state_handler.animations.items()},
        #     parent=self, position=window.left, origin=(-.5,0),)
        # self.side_menu = Entity(parent=self, position=window.left)
        # for key, value in self.state_handler.animations.items():
        #     MenuButton(parent=self.side_menu, text=key.title(), on_click=Func(setattr, self.state_handler, 'state', key), origin_x=-.5)
        # grid_layout(self.side_menu.children, max_x=1, spacing=(0,.01), origin=(-.5,0))
        tabs = ButtonGroup([f'{e.title(): ^24}' for e in self.state_handler.animations.keys()], origin=(0,0), y=.45)
        def on_tab_changed():
            self.state_handler.state = tabs.value.strip().lower()
        tabs.on_value_changed = on_tab_changed
        # # options_back = MenuButton(parent=options_menu, text='Back', x=-.25, origin_x=-.5, on_click=Func(setattr, state_handler, 'state', 'main_menu'))



if __name__ == '__main__':
    from ursina import Ursina, window, Button
    app = Ursina()
    # Text.default_font = 'VeraMono.ttf'
    window.color = hsv(0, 0, 10/255)
    Button.default_color = color._24
    Button.default_highlight_color = color._32
    # ButtonGroup.default_selected_color = color.hex('#85a9f7')
    # ButtonGroup.default_selected_color = color.hex('#ff5eef')
    # ButtonGroup.default_selected_color = color.hex('#85a9f7')
    # color.text_color = hsv(240, 1, .8)
    # color.text_color = hsv(60, 0, 10/255)
    # Button.default_color
    options_menu = OptionsMenu()
    # window.color = color._16
    background = Entity(parent=camera.ui, model='quad', texture='shore', scale=(camera.aspect_ratio,1), color=color.dark_gray, z=1, world_y=0)
    # Audio('chillstep_1.ogg', loop=True, group='music')
    app.run()



# How to move forward while
# taking rotation into account

from ursina import *
from ursina.shaders import lit_with_shadows_shader #!hide
Entity.default_shader = lit_with_shadows_shader #!hide

app = Ursina()
window.size = Vec2(1080, 1920) * .5 #!hide
window.position = (0,0) #!hide

# Make an Entity for the car
car = Entity(
  model='car',
  color=color.red,
  z=-3,
  speed=8
  )
car.model.save('car.bam')

# Make the car rotate
def update():
  car.rotation_y += (
    (held_keys['d']-held_keys['a'])
# Multiply with time.dt to make
# it framerate independent
    * time.dt
# It's a little slow, so let's
# multiply it by 100
    * 100
    )

# Changing car.z won't
# take rotation into account.
# Do this instead:
  car.position += (
# Get the input.
# w to move forward,
# s to move backward
    (held_keys['w']-held_keys['s'])
    * car.speed * time.dt
# Get forward vector with .forward
# Facing straight: Vec3(0,0,1).
# Facing right: Vec3(1,0,0).
# And so on.
    * car.forward
    )

Entity(model='plane', scale=40, texture='grass', texture_scale=Vec2(5)) #!hide
camera.z = -15 #!hide
EditorCamera(rotation=(30,-10,0)) #!hide
DirectionalLight().look_at(Vec3(.5,-1,.25)) #!hide
Sky() #!hide
# Run
app.run()



from ursina import *; hqx_shader = Shader(language=Shader.GLSL, fragment='''
#version 130
// const float SEG = 40.0 / 256.0;
const float SEG = 1.0;

uniform sampler2D p3d_Texture0;
in vec2 uv;
in float iTime;
uniform vec2 texture_size;
out vec4 fragColor;

void main() {
	float mx = 1.0; // start smoothing wt.
 	const float k = -1.10; // wt. decrease factor
 	const float max_w = 0.75; // max filter weigth
 	const float min_w = 0.03; // min filter weigth
 	const float lum_add = 0.33; // effects smoothing

	vec4 color = texture(p3d_Texture0, uv);
    vec3 c = color.xyz;




    float x = 1.0 / (texture_size.x * 2.0);
    float y = 1.0 / (texture_size.y * 2.0);

    const vec3 dt = 1.0*vec3(1.0, 1.0, 1.0);

    vec2 dg1 = vec2( x, y);
    vec2 dg2 = vec2(-x, y);



    vec2 sd1 = dg1 / 2.0;
    vec2 sd2 = dg2 / 2.0;

    vec2 ddx = vec2(x, 0.0);
    vec2 ddy = vec2(0.0, y);

    vec4 t1 = vec4(uv-sd1, uv-ddy);
    vec4 t2 = vec4(uv-sd2, uv+ddx);
    vec4 t3 = vec4(uv+sd1, uv+ddy);
    vec4 t4 = vec4(uv+sd2, uv-ddx);
    vec4 t5 = vec4(uv-dg1, uv-dg2);
    vec4 t6 = vec4(uv+dg1, uv+dg2);

    vec3 i1 = texture(p3d_Texture0, t1.xy).xyz;
    vec3 i2 = texture(p3d_Texture0, t2.xy).xyz;
    vec3 i3 = texture(p3d_Texture0, t3.xy).xyz;
    vec3 i4 = texture(p3d_Texture0, t4.xy).xyz;

    vec3 o1 = texture(p3d_Texture0, t5.xy).xyz;
    vec3 o3 = texture(p3d_Texture0, t6.xy).xyz;
    vec3 o2 = texture(p3d_Texture0, t5.zw).xyz;
    vec3 o4 = texture(p3d_Texture0, t6.zw).xyz;

    vec3 s1 = texture(p3d_Texture0, t1.zw).xyz;
    vec3 s2 = texture(p3d_Texture0, t2.zw).xyz;
    vec3 s3 = texture(p3d_Texture0, t3.zw).xyz;
    vec3 s4 = texture(p3d_Texture0, t4.zw).xyz;

    float ko1 = dot(abs(o1-c),dt);
    float ko2 = dot(abs(o2-c),dt);
    float ko3 = dot(abs(o3-c),dt);
    float ko4 = dot(abs(o4-c),dt);

    float k1=min(dot(abs(i1-i3),dt),max(ko1,ko3));
    float k2=min(dot(abs(i2-i4),dt),max(ko2,ko4));

    float w1 = k2; if(ko3<ko1) w1*=ko3/ko1;
    float w2 = k1; if(ko4<ko2) w2*=ko4/ko2;
    float w3 = k2; if(ko1<ko3) w3*=ko1/ko3;
    float w4 = k1; if(ko2<ko4) w4*=ko2/ko4;

    c = (w1*o1+w2*o2+w3*o3+w4*o4+0.001*c)/(w1+w2+w3+w4+0.001);
    w1 = k*dot(abs(i1-c)+abs(i3-c),dt)/(0.125*dot(i1+i3,dt)+lum_add);
    w2 = k*dot(abs(i2-c)+abs(i4-c),dt)/(0.125*dot(i2+i4,dt)+lum_add);
    w3 = k*dot(abs(s1-c)+abs(s3-c),dt)/(0.125*dot(s1+s3,dt)+lum_add);
    w4 = k*dot(abs(s2-c)+abs(s4-c),dt)/(0.125*dot(s2+s4,dt)+lum_add);

    w1 = clamp(w1+mx,min_w,max_w);
    w2 = clamp(w2+mx,min_w,max_w);
    w3 = clamp(w3+mx,min_w,max_w);
    w4 = clamp(w4+mx,min_w,max_w);

	color = vec4((w1*(i1+i3)+w2*(i2+i4)+w3*(s1+s3)+w4*(s2+s4)+c)/(2.0*(w1+w2+w3+w4)+1.0), 1.0);
	fragColor = color; //texture(iChannel0, uv);
}
''',
)



if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina(size=(1920,1080), position=(0,0))
    window.color=color.black

    shader = hqx_shader
    # Texture.default_filtering = 'bilinear'
    e = Sprite(shader=shader, texture='arrow_down')
    e.set_shader_input('texture_size', e.texture.size)
    EditorCamera()

    e = Sprite(x=-.65, texture='arrow_down')


    def input(key):
        if key == 'space':
            if e.shader == shader:
                e.shader = None
            else:
                e.shader = shader

    from ursina.shaders import fxaa_shader
    camera.shader = fxaa_shader

    app.run()



from ursina import *
from copy import copy



app = Ursina()

dummy = Entity(model='cube')

cube = Entity(model='cube')
cube.model.colorize()
cube = cube.model

ground_plane = Entity(model='plane', scale=32, origin=(-.5,0,-.5), position=(-.5,0,-.5), collider='box', color=color.azure)
cursor = Entity(model=Cube(mode='line'), origin_y=-.5, scale_y=8, color=color.red)
level = Entity(model=Mesh())
i = 0


for z in range(32):
    for x in range(32):
        level.model.vertices += [Vec3(*v) + Vec3(x,0,z) for v in cube.vertices]
        level.model.triangles += [v for v in cube.triangles]
        level.model.colors += [v for v in cube.colors]

level.model.generate()


def update():
    if mouse.hovered_entity == ground_plane:
        cursor.world_position = mouse.world_point
        cursor.world_x = int(cursor.world_x)
        cursor.world_z = int(cursor.world_z)


def on_click():
    x, y = int(cursor.world_x), int(cursor.world_z)
    print(x,y)

ground_plane.on_click = on_click

EditorCamera()
app.run()



import importlib
import importlib.util
import sys
from pathlib import Path
from ursina.string_utilities import print_warning




class CythonFallbackImporter:
    def __init__(self, auto_compile=False):
        self.auto_compile = auto_compile  # store the auto_compile option
        # print('----------------auto_compile', self.auto_compile)
        try:
            from Cython.Build import cythonize
            self.cythonize = cythonize
        except ImportError:
            self.cythonize = None

        print('has cython:', self.cythonize)


    def find_spec(self, fullname, path=None, target=None):
        # if not fullname.endswith('.py'):
        #     print('skip', fullname)
        #     return None  # only handle .py files
        

        # print(f"Trying to load module: {fullname}")
        if path is None:
            path = sys.path  # use default sys.path if no path provided

        path = [Path(p) for p in path]
        
        local_paths = [p for p in path if 'site-packages' not in p.parts and 'lib' not in p.parts]
        # print('aaaaaaaaaa', path)
        # return
        compiled_filename = f"{fullname.replace('.', '/')}.cpython-{sys.version_info[0]}{sys.version_info[1]}.so"  # compiled file name
        pure_python_filename = f"{fullname.replace('.', '/')}.py"  # pure python file name
        pyx_filename = f"{fullname.replace('.', '/')}.apyx"  # .pyx file name for Cython

        # check for the compiled file first
        for directory in local_paths:
            compiled_path = directory / compiled_filename  # use Path directly
            if compiled_path.exists():
                print('import compiled:', fullname)
                return importlib.util.spec_from_file_location(fullname, str(compiled_path))

        # if not found and auto_compile is enabled, try to compile the .pyx file
        if self.auto_compile and self.cythonize is not None:
            for directory in local_paths:
                python_path = directory / pure_python_filename
                pyx_path = directory / pyx_filename  # construct path to .pyx file
                # if any(ignored in python_path.parts for ignored in ['Cython', 'site-packages']):
                #     print('ignore:', python_path)
                #     break
                
                print('try to auto compile:', python_path)
                # if pyx_path.exists():
                with pyx_path.open('w') as destination:
                    with open(directory / pure_python_filename, 'r') as source:
                        destination.write(source.read())
                    print(';;;;;;;;;;;;;create', pyx_path, 'from', directory / pure_python_filename)
                    try:
                        print(f"Compiling {pyx_path} with Cython...")
                        self.cythonize(pyx_path, force=True)
                        compiled_filename = f"{module_name}.cpython-{sys.version_info[0]}{sys.version_info[1]}.so"
                        print(f"successfully compiled {compiled_filename}")
                        compiled_path = Path(directory) / compiled_filename
                        return importlib.util.spec_from_file_location(fullname[:-4], str(compiled_path))

                    except Exception as e:
                        print_warning(f"Failed to compile {pyx_path} with Cython: {e}")
                        break

        # fallback to pure Python file if no compiled file is found
        for directory in path:
            python_path = directory / pure_python_filename  # use Path directly
            if python_path.exists():
                print_warning(f'import cython module failed, falling back to python for: {Path(python_path).name}')
                return importlib.util.spec_from_file_location(fullname, str(python_path))

        print('failed import')
        return None


def replace_import_statement(auto_compile=False):
    cython_importer = CythonFallbackImporter(auto_compile=auto_compile)
    # print('---------', cython_importer.pyximport)
    if auto_compile and cython_importer.cythonize is None:
        print_warning('''trying to use auto_compile, but pyximport wasn't found (already compiled modules can still be imported)''')

    sys.meta_path.insert(0, cython_importer)  # register custom importer


if __name__ == '__main__':
    replace_import_statement(auto_compile=True)  # change this to True to enable auto-compilation
    import grid_layout
    print('----', grid_layout)
    # from ursina.scripts.grid_layout import grid_layout  # test the import



from ursina import *


class RadialMenu(Entity):
    def __init__(self, buttons=list(), **kwargs):
        super().__init__()
        self.parent = camera.ui
        self.buttons = buttons
        origin = Entity(parent=self)
        self.open_at_cursor = True
        self.open_duration = .1

        self.bg = Panel(parent=self, model='quad', z=99, scale=999, collider='box', color=color.hsv(0,0,0,.1), enabled=False)
        self.z = -99

        offset = lerp(.5, 2, len(self.buttons)/8)
        for i, b in enumerate(self.buttons):
            b.parent = origin
            b.y = offset
            origin.rotation_z = i/len(self.buttons) * 360
            b.world_parent = self
            b.rotation_z = 0
            if hasattr(b, 'text_entity') and b.text_entity:
                b.text_entity.world_scale /= .075

        destroy(origin)
        self.scale = .075

        for key, value in kwargs.items():
            setattr(self, key, value)



    def on_enable(self):
        if not hasattr(self, 'bg'):
            return

        self.bg.enabled = True
        if self.open_at_cursor:
            self.position = mouse.position

        delay_step = self.open_duration / (len(self.children)-1)
        original_scales = [c.scale for c in self.children]
        for i, c in enumerate(self.children):
            if c is self.bg:
                continue

            c.scale = 0
            c.animate_scale(original_scales[i], duration=self.open_duration, delay=i*delay_step, curve=curve.linear)


    def input(self, key):
        if key == 'left mouse down' and mouse.hovered_entity in [c for c in self.children if isinstance(c, Button)]:
            invoke(setattr, self, 'enabled', False, delay=.1)
        elif key == 'left mouse down' and mouse.hovered_entity == self.bg:
            invoke(setattr, self, 'enabled', False, delay=.1)


class RadialMenuButton(Button):
    def __init__(self, **kwargs):
        super().__init__(model='sphere', scale=1.3, highlight_scale=1.2, pressed_color=color.azure)


        for key, value in kwargs.items():
            setattr(self, key, value)



if __name__ == '__main__':
    app = Ursina()

    rm = RadialMenu(
        buttons = (
            RadialMenuButton(text='1'),
            RadialMenuButton(text='2'),
            RadialMenuButton(text='3'),
            RadialMenuButton(text='4'),
            RadialMenuButton(text='5', scale=.5),
            RadialMenuButton(text='6', color=color.red),
            ),
        enabled = False
        )
    RadialMenuButton(text='6', color=color.red,x =-.5, scale=.06),
    def enable_radial_menu():
        rm.enabled = True
    cube = Button(parent=scene, model='cube', color=color.orange, highlight_color=color.azure, on_click=enable_radial_menu)
    EditorCamera()
    app.run()



from ursina import *

app = Ursina()

window.color = color.black
print(len(scene.entities))
base_cube = Entity(model='cube', color=color.yellow, texture='white_cube')

for i in range(8):
    print(curve.out_expo(i/8))
    Entity(parent=base_cube, model='cube', color=color.orange, scale=math.pow(1.14,i), alpha=.1)

base_cube.combine()
print(len(scene.entities))


combine_parent = Entity()
base_cube.parent = combine_parent

# for i in range(20):
#     duplicate(base_cube, position=(random.uniform(-100,100),random.uniform(-100,100),random.uniform(-100,100)))

# combine_parent.combine()

EditorCamera()


app.run()



from ursina import *;

def randVec(i, n):
  """Generates a random point in a hemisphere."""

  x = random.uniform(-1, 1)
  y = random.uniform(-1, 1)
  z = random.uniform( 0.5, 1)

  l = random.uniform(0.1, 1) / math.sqrt(x * x + y * y + z * z)

  scale = (i + 1.0) / n
  l *= 0.1 + 0.9 * scale * scale

  return (x * l, y * l, z * l)


VECS = [randVec(i, 8) for i in range(8)]

ssao_shader = Shader(language=Shader.GLSL,
vertex = '''
#version 410 core
layout(location = 0) in vec3 i_vertex;
out vec2 v_uv;
void main() {
    v_uv = (i_vertex.xy + 1.0) / 2.0;
    gl_Position = vec4(i_vertex, 1.0);
}
''',

fragment='''
#version 410 core

const vec3 SAMPLES[] = vec3[](%(samples)s);
const float FOCUS = 0.15f;
const float POWER = 5.0;
// uniform mat4 u_proj;
uniform mat4 p3d_ProjectionMatrixInverse;
uniform sampler2D u_random;
uniform sampler2D ntex;
// uniform sampler2D u_vertex;
layout(location = 0) out float ao;
in vec2 v_uv;
in sampler2D d_tex;

void main() {
  // Read the position, normal and the random vector.
  vec2 scale = textureSize(ntex, 0) / textureSize(u_random, 0);
  vec4 vert = texture(u_vertex, v_uv);
  vec3 norm = texture(ntex, v_uv).rgb;
  // vec3 rvec = texture(u_random, v_uv * scale).rgb;
  // Compute the btn matrix to rotate the hemisphere.
  vec3 tang = normalize(rvec - norm * dot(rvec, norm));
  vec3 btan = cross(norm, tang);
  mat3 tbn = mat3(tang, btan, norm);
  float occlusion = 0.0;
  for (int i = 0; i < %(count)d; ++i) {
    // Compute an offset from the center point & project it.
    vec3 smpl_view = tbn * SAMPLES[i] * FOCUS + vert.xyz;
    vec4 smpl_proj = p3d_ProjectionMatrixInverse * vec4(smpl_view, 1);
    smpl_proj.xyz /= smpl_proj.w;
    smpl_proj.xy = smpl_proj.xy * 0.5f + 0.5f;
    // Sample the depth.
    float depth = -texture(u_vertex, smpl_proj.xy).w;
    // Check for edges with a steep falloff.
    float range = smoothstep(0, 1, FOCUS / abs(vert.w - depth));

    // If point occluded, add it to the accumulator.
    occlusion += depth > smpl_view.z ? 1 : 0;
  }
  ao = pow(1.0 - occlusion / %(count)d, POWER);
}
''' % {
  'samples': ','.join(['vec3(%1.4f,%1.4f,%1.4f)' % v for v in VECS]),
  'count': len(VECS)
}


,
default_input = {
    # 'random_texture' : Func(load_texture, 'noise')
    'u_random' : Func(load_texture, 'noise')
}
)


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere', color=color.orange)
    e = Entity(model='cube', y=-1)
    e = Entity(model='plane', scale=100, y=-1)
    camera.clip_plane_far = 100
    camera.shader = ssao_shader
    EditorCamera()

    def input(key):
        if key == 'space':
            if camera.shader:
                camera.shader = None
            else:
                camera.shader = ssao_shader

    app.run()



from ursina import input_handler, held_keys, Entity
from panda3d.core import InputDevice, InputDeviceManager

main_gamepad_index = 0


if __name__ == '__main__':
    from ursina import Ursina, Text, time, color, window
    app = Ursina()
    window.color = color.black


# input_handler.gamepad = None
input_handler.gamepads = base.devices.getDevices(InputDevice.DeviceClass.gamepad)

# def disconnect_all():

def connect_all():
    for i, gamepad in enumerate(input_handler.gamepads):
        try:
            base.detachInputDevice(gamepad)
        except:
            pass

    for i, gamepad in enumerate(input_handler.gamepads):
        gamepad_name = 'gamepad'
        if i != main_gamepad_index:
            gamepad_name += f'_{i}'

        base.attachInputDevice(gamepad, prefix=gamepad_name)
        buttons = {
            'face_a' : 'a',
            'face_b' : 'b',
            'face_x' : 'x',
            'face_y' : 'y',
            'lstick' : 'left stick',
            'rstick' : 'right stick',
            'back' : 'back',
            'start' : 'start',
            'dpad_up' : 'dpad up',
            'dpad_down' : 'dpad down',
            'dpad_left' : 'dpad left',
            'dpad_right' : 'dpad right',
            'rshoulder' : 'right shoulder',
            'lshoulder' : 'left shoulder',
        }

        for original_name, new_name in buttons.items():
            base.accept(f'{gamepad_name}-{original_name}', base.input, extraArgs=[f'{gamepad_name} {new_name}'])
            base.accept(f'{gamepad_name}-{original_name}-up', base.input, extraArgs=[f'{gamepad_name} {new_name} up'])
            # print(original_name, new_name)

def update():
    for i, gamepad in enumerate(input_handler.gamepads):
        gamepad_name = 'gamepad'
        if i != main_gamepad_index:
            gamepad_name += f'_{i}'

        # left stick
        x = gamepad.findAxis(InputDevice.Axis.left_x).value
        if abs(x) < .1:
            x = 0
        held_keys[f'{gamepad_name} left stick x'] = x

        y = gamepad.findAxis(InputDevice.Axis.left_y).value
        if abs(y) < .1:
            y = 0
        held_keys[f'{gamepad_name} left stick y'] = y

        # right stick
        x = gamepad.findAxis(InputDevice.Axis.right_x).value
        if abs(x) < .1:
            x = 0
        held_keys[f'{gamepad_name} right stick x'] = x

        y = gamepad.findAxis(InputDevice.Axis.right_y).value
        if abs(y) < .1:
            y = 0
        held_keys[f'{gamepad_name} right stick y'] = y

        held_keys[f'{gamepad_name} left trigger'] = gamepad.findAxis(InputDevice.Axis.left_trigger).value
        held_keys[f'{gamepad_name} right trigger'] = gamepad.findAxis(InputDevice.Axis.right_trigger).value

from ursina import curve, invoke, Func
def vibrate(gamepad_index=0, low_freq_motor=1, high_freq_motor=1, duration=.1, curve=curve.one):
    if not input_handler.gamepads or gamepad_index < len(input_handler.gamepads)-1:
        return
    input_handler.gamepads[gamepad_index].set_vibration(low_freq_motor, high_freq_motor)
    invoke(Func(input_handler.gamepads[gamepad_index].set_vibration, 0, 0), delay=duration)

Entity(name='gamepad_handler', update=update, eternal=True) # connect update() to an entity so it runs
connect_all()


if __name__ == '__main__':
    text_entity = Text()
    print('---------------', input_handler.gamepads)
    [print('-----------', gamepad) for gamepad in input_handler.gamepads]
 
    gamepad_list = Text('\n'.join([f'{i}) {gamepad}' for i, gamepad in enumerate(input_handler.gamepads)]), x=-.5, y=.3)
    player = Entity(model='cube', color=color.azure)

    def update():
        player.x += held_keys['gamepad left stick x'] * time.dt * 5
        player.y += held_keys['gamepad left stick y'] * time.dt * 5
        text_entity.text = '\n'.join([f'{key}: {held_keys[key]}' for key in ('gamepad left trigger', 'gamepad right trigger', 'gamepad right stick x', 'gamepad right stick y')])

    def input(key):
        global main_gamepad_index
        print('key:', key)
        if key.isdigit():
            i = int(key)
            if i < len(input_handler.gamepads):
                print('set main gamepad to:', input_handler.gamepads[i])
                main_gamepad_index = i
                connect_all()

        if key == 'gamepad x':
            from ursina import gamepad
            gamepad.vibrate()
                # input_handler.gamepads = [target_gamepad, ] + input_handler.gamepads
                # target_gamepad = input_handler.gamepads.pop(i)
                # input_handler.gamepad = input_handler.gamepads[i]

    app.run()



from ursina import Shader; my_shader = Shader(fragment='''
#version 430

uniform sampler2D tex;
uniform sampler2D tex2;
in vec2 uv;
out vec4 color;
in vec2 window_size;

uniform bool on;
uniform float blur_size;
// #define img_size iResolution.xy

vec4 blur(sampler2D tex, vec2 uv, vec2 direction) {
    vec4 col = vec4(0.);

    for(float index=0; index<10; index++) {
        vec2 offset_uv = uv + vec2(
            (index/9 - 0.5) * blur_size * direction.x,
            (index/9 - 0.5) * blur_size * direction.y);
        col += texture(tex, offset_uv);
    }

    col = col / 10;
    texture(tex, uv);
    return col;
}

void main() {
    vec4 rgba = texture(tex, uv).rgba;
    vec4 vertical = blur(tex, uv, vec2(0.,1.));
    vec4 horizontal = blur(tex2, uv, vec2(1.,0.));

    color = (vertical*.5) + (horizontal*.5);

}

''',
default_input = dict(
    on = True,
    blur_size = .1
)
)



if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    buffer = base.win.makeTextureBuffer("My Buffer", *[int(e*1) for e in window.size])
    tex2 = buffer.getTexture()
    my_shader.default_input['tex2'] = tex2

    mycamera = base.makeCamera(buffer)
    mycamera.node().set_lens(camera.lens)
    mycamera.reparentTo(camera)


    e = Entity(model='sphere', color=color.yellow)
    e = Entity(model='cube', y=-1)
    camera.shader = my_shader

    settings = Empty(on=camera.shader.default_input['on'])

    def input(key):
        if key == 'space':
            settings.on = not settings.on
            camera.set_shader_input('on', settings.on)

        if key == '6':
            window.vsync = 60
        if key == '1':
            window.vsync = 100

    EditorCamera()
    window.vsync = 100

    app.run()



from ursina import *    #import everything we need with just one line


class Player(Entity):   # inherits Entity, out base class for most things. Like GameObject in Unity
    def __init__(self):
        super().__init__()

        self.position = (0,1,0)
        self.position = (0,1)   # third axis is optional
        self.x = 0
        self.y = 1
        self.z = 0
        self.world_position = (0,1,0)
        self.world_x = 0

        self.rotation = (0,0,0)   # euler angles
        # target = (2,0,1)    # can also be an Entity
        # self.look_at(target)

        self.scale = (1,2,1)
        self.scale_x = 1
        self.scale_y = 2
        self.scale_z = 1
        self.world_scale = (1,2)

        self.model = 'quad' # tries to load model from /models folder
        # you can also set it to a built in procedural model
        # self.model = Circle(resolution=16)

        self.origin = (-.5, -.5)    # sets model offset. often used when working with ui

        # tries to load texture from /textures/compressed folder
        # call compress_textures() to compress .psd files and big png files
        self.texture = 'white_cube'

        self.color = color.hsv(90, 1, 1, 1) #hsv color
        self.collider = 'box'
        self.text_entity = Text(text = 'a     d', parent=scene, z=-.1)

    def input(self, key):
        print(key)

    def update(self):
        self.x += held_keys['d'] * .1
        self.x -= held_keys['a'] * .1

    def on_mouse_enter(self):
        print('enter')

    def on_mouse_exit(self):
        print('exit')

    def on_click(self):
        print('click')




if __name__ == '__main__':
    app = Ursina()
    sky = Sky()
    player = Player()

    button = Button(position=(-.5, .25), text='button_text')
    button.scale *= .2
    button.text_entity.scale /= .2
    button.tooltip = Tooltip('tooltip text')

    info_text = Text(
        '''<black>This is just an <orange>example game \n<black>so it doesn't look that <red>great''',
        y = -3,
        )

    draggable = Draggable(
        scale = (.2,.1),
        texture = 'white_cube',
        color = color.tint(color.lime, -.5),
        text = 'drag & drop!',
        # world_parent = scene
        )

    app.run()



from ursina import *

app = Ursina()

Entity(model='plane', color=color.salmon, scale=10, collider='box')

cursor = Entity(model=Circle(24, mode='line', thickness=4), rotation_x=90)

def input(key):
    if key == 'left mouse down':
        cursor.position = mouse.world_point
        cursor.y += .1
        cursor.scale = 0
        cursor.animate('scale', Vec3(1,1,1), duration=.2, curve=curve.in_out_circ)
        cursor.color = color.white
        cursor.animate('color', color.clear, duration=.2-.05, delay=.05, curve=curve.in_out_circ)




EditorCamera()


app.run()



from ursina import Entity, camera, Text, Vec2, mouse, color, floor, clamp, time, held_keys, destroy, curve
import pyperclip

from ursina.string_utilities import multireplace
# from tree_view import TreeView


class TextField(Entity):
    def __init__(self, max_lines=64, line_height=1.1, character_limit=None, **kwargs):
        super().__init__(parent=camera.ui, x=-.5, y=.4, ignore_paused=True)

        self.font = Text.default_monospace_font
        self.line_height = line_height
        self.max_lines = max_lines
        self.character_limit = character_limit

        self.scroll_parent = Entity(parent=self)
        self.text_entity = Text(parent=self.scroll_parent, start_tag='☾', end_tag='☽', font=self.font, text='', line_height=self.line_height, origin=(-.5, .5))
        self.line_numbers = Text(parent=self.scroll_parent, font=self.font, line_height=line_height, text='0', origin=(.5,.5), x=-.04, color=color.gray, enabled=False)
        self.character_width = Text.get_width('a', font=self.font)
        self.cursor_parent = Entity(parent=self.scroll_parent, scale=(self.character_width, -1*Text.size*self.line_height))
        self.cursor = Entity(name='text_field_cursor', parent=self.cursor_parent, model='cube', color=color.cyan, origin=(-.5, -.5), scale=(.1, 1, 0), enabled=False)
        self.cursor.animate_color(color.red, duration=1.2, curve=curve.linear_boomerang, loop=True)
        self.bg = Entity(name='text_field_bg', parent=self, model='quad', double_sided=True, color=color.dark_gray, origin=(-.5,.5), z=0.005, scale=(120, Text.size*self.max_lines*self.line_height), collider='box', visible=True)

        self.selection = [Vec2(0,0), Vec2(0,0)]
        self.selection_parent = Entity(name='text_field_selection_parent', parent=self.cursor_parent, scale=(1,1,0))
        self.register_mouse_input = False
        self.world_space_mouse = False

        self.triple_click_delay = 0.3
        self._last_double_click = 0
        self.scroll = 0
        self.scroll_amount = 2
        self._prev_scroll = self.scroll

        self.active = True
        self.highlight_color = color.hsv(120,1,1,.1)
        self.text = ''
        self.delimiters = ' .,!?;:(){}[]<>\'\"@#$%^&*+=-\\|/`~'
        self.replacements = dict()
        self.on_undo = []
        self.on_redo = []
        self.on_value_changed = None

        self.shortcuts = {
            'newline':          ('enter', 'enter hold'),
            'erase':            ('backspace', 'backspace hold'),
            'erase_word':       ('ctrl+backspace', 'ctrl+backspace hold'),
            'delete_line':      ('ctrl+shift+k',),
            'duplicate_line':   ('ctrl+shift+d',),
            'undo':             ('ctrl+z', 'ctrl+z hold'),
            'redo':             ('ctrl+y', 'ctrl+y hold', 'ctrl+shift+z', 'ctrl+shift+z hold'),
            'indent':           ('tab',),
            'dedent':           ('shift+tab',),
            'move_line_down':   ('ctrl+down arrow', 'ctrl+down arrow hold'),
            'move_line_up':     ('ctrl+up arrow', 'ctrl+up arrow hold'),
            'scroll_up':        ('scroll up',),
            'scroll_down':      ('scroll down',),
            'cut':              ('ctrl+x',),
            'copy':             ('ctrl+c',),
            'paste':            ('ctrl+v',),
            'select_all':       ('ctrl+a',),
            'select_word':      ('double click',),
            'select_line':      ('triple click',),
            'scroll_to_bottom': ('shift+alt+e',),
            'move_operations' : {
                'move_left':                ('left arrow', 'left arrow hold', 'shift+left arrow', 'shift+left arrow hold'),
                'move_right':               ('right arrow', 'right arrow hold', 'shift+right arrow', 'shift+right arrow hold'),
                'move_up':                  ('up arrow', 'up arrow hold', 'shift+up arrow', 'shift+up arrow hold'),
                'move_down':                ('down arrow', 'down arrow hold', 'shift+down arrow', 'shift+down arrow hold'),
                'move_to_end_of_word' :     ('ctrl+right arrow', 'ctrl+right arrow hold', 'ctrl+shift+right arrow', 'ctrl+shift+right arrow hold'),
                'move_to_start_of_word' :   ('ctrl+left arrow', 'ctrl+left arrow hold', 'ctrl+shift+left arrow', 'ctrl+shift+left arrow hold'),
            },
            # 'save':             ('ctrl+s',),
            # 'save_as':          ('ctrl+shift+s',),
            # 'toggle_comment':   ('ctrl+alt+c',),
            # 'find':             ('ctrl+f',),
            # 'select_word_left': ('ctrl+shift+left arrow', 'ctrl+shift+left arrow hold'),
            # 'select_word_right': ('ctrl+shift+right arrow', 'ctrl+shift+right arrow hold'),
        }
        def middle_click_input(key):
            if self.register_mouse_input:
                if key == 'middle mouse down':
                    self.middle_click_scroller.start_y = mouse.y
                    self._original_scroll_amount = self.scroll_amount
                elif key == 'middle mouse up' and hasattr(self, '_original_scroll_amount'):
                    self.middle_click_scroller.start_y = None
                    self.scroll_amount = self._original_scroll_amount

        def middle_click_update():
            if not mouse.middle or self.middle_click_scroller.start_y == None:
                return
            self.middle_click_scroller.t += time.dt
            if self.middle_click_scroller.t < self.middle_click_scroller.update_rate:
                return

            # self.middle_click_scroller.update_rate = 1-mouse.delta_drag[0]
            # print('-aaaaaaaaaaa', 1-mouse.delta_drag[0])
            dist = abs(mouse.y - self.middle_click_scroller.start_y)
            self.scroll_amount = 1
            if dist > .1:
                self.scroll_amount = 4
            elif dist > .2:
                self.scroll_amount = 6
            # elif dist > .4:
            #     self.scroll_amount = 12

            self.middle_click_scroller.t = 0

            if self.register_mouse_input:
                if mouse.y < self.middle_click_scroller.start_y:
                    self.input('scroll down')
                if mouse.y > self.middle_click_scroller.start_y:
                    self.input('scroll up')

        self.middle_click_scroller = Entity(parent=self, start_y=None, input=middle_click_input, update=middle_click_update, t=0, update_rate=.05)


        for key, value in kwargs.items():
            setattr(self, key, value)

        self._prev_text = ''



    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, value):
        self._active = value
        self.cursor.enabled = value

        if not value:
            self.selection = [Vec2(0,0), Vec2(0,0)]
            self.draw_selection()


    def add_text(self, s, move_cursor=True, rerender=True):
        if self.character_limit and len(self.text) >= self.character_limit:
            return

        x, y = int(self.cursor.x), int(self.cursor.y)

        lines = self.text.split('\n')
        l = lines[y]
        lines[y] =  l[:x] + s + l[x:]
        self._append_undo(self.text, y, x)
        self.text = '\n'.join(lines)

        if move_cursor:
            self.cursor.x += len(s)

        if rerender:
            self.render()

    def _append_undo(self, text, y, x, clear_redo = True):
        if clear_redo:
            self.on_redo.clear()
        self.on_undo.append((text, y, x))


    def move_line(self, line_index, delta, move_cursor=True):
        x, y = int(self.cursor.x), int(self.cursor.y)

        start_y, end_y = self.cursor.Y, self.cursor.Y

        if self.selection != [Vec2(0,0), Vec2(0,0)]:
            start_y = int(self.selection[0].y)
            end_y = int(self.selection[1].y)

        lines = self.text.split('\n')
        # lines[line_index], lines[line_index+delta] = lines[line_index+delta], lines[line_index]
        # start = []
        # if start_y-1 > 0:

        #     start = lines[:int(start_y-1)]

        middle = lines[start_y:end_y+1]
        del lines[start_y:end_y+1]
        for l in middle[::delta]:
            lines.insert(start_y+delta, l)
        # end = lines[start_y+1:]

        # [print(e) for e in middle]
        self.cursor.y += delta * move_cursor
        for e in self.selection:
            e.y += delta
        self.draw_selection()

        # if delta == 1:
        #     _ = start.pop()
        #     end.insert(0, _)
        # elif delta == -1:
        #     _ = end.pop(0)
        #     start.append(_)
        #
        # lines = start + middle + end
        #
        self._append_undo(self.text, y, x)
        self.text = '\n'.join(lines)
        # print('moved line')


    def erase(self, rerender=True):
        # if not self.selection or self.selection[0] == self.selection[1]:
        x, y = int(self.cursor.x), int(self.cursor.y)
        if x+y == 0:
            return

        lines = self.text.split('\n')
        l = lines[y]
        # delete \n and go to line above
        if x == 0 and y > 0:
            new_x = len(lines[y-1])
            lines[y-1] += l
            lines.pop(y)
            self.cursor.x = new_x
            self.cursor.y -= 1
        # delete tab
        elif l[:x].lstrip() == '' and x >=4:
            l = l[4:]
            self.cursor.x -= 4
            lines[y] = l
        # normal erase
        else:
            l = l[:x-1] + l[x:]
            self.cursor.x -= 1

            lines[y] = l

        self._append_undo(self.text, y, x)

        # else:   # delete selected
        #     self.delete_selected()
        #     lines = self.text.split('\n')
        #
        #     start_y = int(self.selection[0][1])
        #     end_y = int(self.selection[1][1])
        #
        #     if start_y == end_y:    # one line
        #         l = lines[start_y]
        #         start_x, end_x = int(self.selection[0][0]), int(self.selection[1][0])
        #         l = l[:start_x] + l[end_x:]
        #         lines[start_y] = l
        #         self.selection = (self.selection[0], self.selection[0])
        #         self.cursor.position = self.selection[0][0]
                # self.selection = None
                # self.selection[0][0],start_y),
                # ignore=True, scale_x=self.selection[1][0]-self.selection[0][0])
            # for y in range(int(self.selection[0][1]), int(self.selection[1][1])):
            #     lines.pop(y)




        self.text = '\n'.join(lines)
        if rerender:
            self.render()

    def _ordered_selection(self):
        if not self.selection:
            return None
        if self.selection[1][1] < self.selection[0][1] or (self.selection[1][1] == self.selection[0][1] and self.selection[1][0] < self.selection[0][0]):
            return [self.selection[1], self.selection[0]]
        return self.selection


    def delete_selected(self):
        if not self.selection or self.selection[0] == self.selection[1]:
            return

        sel = self._ordered_selection()
        self.cursor.position = sel[0]
        start_y = int(sel[0][1])
        end_y = int(sel[1][1])
        lines = self.text.split('\n')

        lines[start_y] = lines[start_y][:int(sel[0][0])] + lines[end_y][int(sel[1][0]):]
        del lines[(start_y + 1) : (end_y + 1)]

        self._append_undo(self.text, self.cursor.y, self.cursor.x)
        self.text = '\n'.join(lines)
        self.selection = [Vec2(0,0), Vec2(0,0)]
        # self._append_undo(self.text, y, x)
        self.draw_selection()
        self.render()


    def get_selected(self):
        if not self.selection or self.selection[0] == self.selection[1]:
            return None

        sel = self._ordered_selection()
        start_y = int(sel[0][1])
        end_y = int(sel[1][1])
        lines = self.text.split('\n')

        selected_text = ''
        # selected_text = lines[start_y][]

        for y in range(start_y, end_y+1):
            if y > start_y:
                selected_text += '\n'
            selected_text += lines[y][(int(sel[0][0]) if y == start_y else 0) : (int(sel[1][0]) if y == end_y else len(lines[y])) ]

        return selected_text


    def get_mouse_position_unclamped(self):
        if self.world_space_mouse:
            if mouse.point:
                x = round(mouse.point[0] * self.bg.scale_x)
                y = floor(mouse.point[1] * self.bg.scale_y)
            else:
                x = int(self.cursor.x)
                y = int(self.cursor.y)

        else:
            mpos = mouse.position
            mpos.x = mpos.x / self.world_scale_x * camera.ui.world_scale_x
            mpos.y = mpos.y / self.world_scale_y * camera.ui.world_scale_y
            mpos += camera.ui.get_position(self.text_entity)

            x = round(mpos.x / self.cursor_parent.scale_x)
            y = floor(mpos.y / self.cursor_parent.scale_y)

        return (x, y)


    def get_mouse_position(self):
        (x, y) = self.get_mouse_position_unclamped()

        lines = self.text.split('\n')
        y = clamp(y, 0, len(lines) - 1)
        x = clamp(x, 0, len(lines[y]))

        return (x, y)

    def set_scroll(self, value, render=True):
        self.scroll = value
        if self.max_lines < 9999:
            if self.scroll < 9999:
                self.scroll = clamp(self.scroll, 0, 9999)
                self.render()
        else:
            self.scroll = clamp(self.scroll, 0, 9999)
            self.scroll_parent.y = (self.scroll * Text.size * self.line_height)

        if render:
            self.render()


    def input(self, key):
        # print('-------------', key)
        text, cursor, on_undo, add_text, erase = self.text, self.cursor, self.on_undo, self.add_text, self.erase

        if self.register_mouse_input and self.bg.hovered and key == 'left mouse up':
            self.active = True

        elif self.register_mouse_input and not self.bg.hovered and key == 'left mouse down':    # clicked outside
            self.active = False

        if not self.active:
            return

        if mouse.hovered_entity == self.bg:
            if key in self.shortcuts['scroll_down']:
                self.set_scroll(self.scroll + self.scroll_amount)

            elif key in self.shortcuts['scroll_up']:
                self.set_scroll(self.scroll - self.scroll_amount)

        if key == 'double click':
            t = time.time()
            if t - self._last_double_click < self.triple_click_delay:
                key = 'triple click'
            self._last_double_click = t

        # key = f'{"ctrl" if held_keys["control"]} + {key}'
        ctrl, shift, alt = '', '', ''
        if held_keys['control'] and key != 'control':
            ctrl = 'ctrl+'
        if held_keys['shift'] and key != 'shift':
            shift = 'shift+'
        if held_keys['alt'] and key != 'alt':
            alt = 'alt+'

        key = ctrl+shift+alt+key
        x, y = int(cursor.x), int(cursor.y)
        lines = text.split('\n')
        l = lines[y]

        if key in self.shortcuts['move_operations']['move_up']:
            if y > 0:
                cursor.y -= 1
                cursor.x = min(x, len(lines[y-1]))
        if key in self.shortcuts['move_operations']['move_down']:
            if y < len(lines)-1:
                cursor.y += 1
                cursor.x = min(x, len(lines[y+1]))
        if key in self.shortcuts['move_operations']['move_right']:
            if x == len(l) and y < len(lines)-1:        # end of line, move to beginning of next
                cursor.y += 1
                cursor.x = 0
            elif x < len(l):
                cursor.x += 1
        if key in self.shortcuts['move_operations']['move_left']:
            if x > 0:
                cursor.x -= 1
            elif y > 0:                                 # move to end of line above
                cursor.y -= 1
                cursor.x = len(lines[y-1])


        delimiters = self.delimiters

        if key in self.shortcuts['move_operations']['move_to_start_of_word']:
            self.move_to_start_of_word()
            if x < len(l) and (l[x] in delimiters or x == 0):
                self.move_to_start_of_word()
            if held_keys['shift']:
                self.selection[1] = self.cursor.position
                self.draw_selection()

        if key in self.shortcuts['move_operations']['move_to_end_of_word']:
            self.move_to_end_of_word()
            if x == len(l) or l[x] in delimiters:
                self.move_to_end_of_word()

        moved_cursor = False
        for move in self.shortcuts['move_operations']:
            if key in self.shortcuts['move_operations'][move]:
                moved_cursor = True
                if not held_keys['shift']:
                    self.selection = [Vec2(self.cursor.X, self.cursor.Y), Vec2(self.cursor.X, self.cursor.Y)]
                    self.draw_selection()
                break

        if moved_cursor and held_keys['shift']:
            self.selection[1] = self.cursor.position
            self.draw_selection()


        if key in self.shortcuts['newline'] and (self.cursor.y < self.max_lines-1 or self.max_lines == 9999):
            if self.selection:
                self.delete_selected()

            if l.startswith('class ') and not l.endswith(':'):
                add_text(':', rerender=False)
            if l.startswith('def ') and not l.endswith(':'):
                add_text(':', rerender=False)

            add_text('\n', rerender=False)

            if l.strip() == '':
                indent = len(l)
            else:
                indent = 0
                if l.lstrip().startswith('class ') or l.lstrip().startswith('def '):
                    indent += 4

            self.cursor.y += 1
            self.cursor.x = 0
            add_text(' '*indent)
            return
            # self.cursor.x = indent

        if key in self.shortcuts['indent']:
            if not self.selection or self.selection[0] == self.selection[1]:
                add_text(' '*4, move_cursor=True)
                return

            else:
                for y in range(int(self.selection[0][1]), int(self.selection[1][1])+1):
                    # print('indent', y)
                    lines[y] = (' '*4) + lines[y]
                self.cursor.x += 4
                self._append_undo(self.text, y, x)
                self.text = '\n'.join(lines)
                self.render()
                return

        if key in self.shortcuts['dedent']:
            if not self.selection or self.selection[0] == self.selection[1]:
                if l.startswith(' '*4):
                    lines[y] = l[4:]
            else:
                for y in range(int(self.selection[0][1]), int(self.selection[1][1])+1):
                    l = lines[y]
                    if l.startswith(' '*4):
                        # print('dedent')
                        lines[y] = l[4:]

            self.cursor.x = max(self.cursor.x - 4, 0)
            self._append_undo(self.text, y, x)
            self.text = '\n'.join(lines)
            self.render()
            return

        if key in self.shortcuts['scroll_to_bottom'] and mouse.hovered_entity == self.bg:
            self.scroll_to_bottom()


        if key in self.shortcuts['erase']:
            if not self.selection or self.selection[1] == self.selection[0]:
                erase()
                return
            else:
                self.delete_selected()
                return

        if key in self.shortcuts['erase_word']:
            # print('delete word')
            if x == 0:
                erase()
                return

            if ' ' not in l:
                l = l[x:]
                self.cursor.x = 0

            else:
                removed = ''
                beginning = l[:x]


                if beginning.endswith('  '): # delete whitespace
                    beginning = beginning.rstrip()
                    removed = ' ' * (x-len(beginning))

                for delimiter in ('.', '\'', '\"', '(', '"', '\''):
                    beginning = beginning.replace(delimiter, ' ')

                if ' ' not in beginning: # first word of the line
                    beginning = ''
                else:
                    beginning = beginning.rstrip().rsplit(' ', 1)[0]

                removed = l[len(beginning):x]

                l = beginning + l[x:]
                self.cursor.x -= len(removed)

            lines[y] = l
            self._append_undo(self.text, y, x)
            self.text = '\n'.join(lines)
            self.render()
            return

        if key in self.shortcuts['move_line_down'] and self.cursor.y < self.max_lines:
            # print('move down')
            if y+1 == len(lines): # if at last line
                self.text += '\n'

            self.move_line(y, 1)
            self.render()
            return

        if key in self.shortcuts['move_line_up'] and y > 0:
            self.move_line(y, -1)
            self.render()
            return

        if key in self.shortcuts['undo']:
            if not on_undo:
                return

            self.on_redo.append((self.text, y, x))
            self.text = on_undo[-1][0]
            cursor.y = on_undo[-1][1]
            cursor.x = on_undo[-1][2]
            on_undo.pop()
            self.render()
            return

        if key in self.shortcuts['redo']:
            if not self.on_redo:
                return

            on_undo.append((self.text, y, x))
            self.text = self.on_redo[-1][0]
            cursor.y = self.on_redo[-1][1]
            cursor.x = self.on_redo[-1][2]
            self.on_redo.pop()
            self.render()
            return

        if key in self.shortcuts['delete_line']:
            self._append_undo(self.text, y, 0)
            lines.pop(y)

            if y == 0:
                self.cursor.x = 0
            else:
                self.cursor.x = len(lines[y-1])

            if y >= len(lines) and y > 0:
                self.cursor.y -= 1

            self.text = '\n'.join(lines)
            self.render()
            return

        if key in self.shortcuts['duplicate_line']:
            if len(lines) < self.max_lines:
                self._append_undo(self.text, y, 0)
                lines.insert(y, lines[y])
                self.text = '\n'.join(lines)
                cursor.y += 1


        if key in self.shortcuts['select_word']:
            if x >= len(l):
                return

            for start_x in range(x, -1, -1):    # go left until end of word
                if l[start_x] in delimiters:
                    break
            if start_x > 0:
                start_x += 1

            for end_x in range(x, len(l)):
                if l[end_x] in delimiters:
                    break
            if end_x == len(l)-1:
                end_x += 1


            self.selection = [Vec2(start_x, y), Vec2(end_x, y)]
            self.draw_selection()
            return

        if key in self.shortcuts['select_line']:
            self.selection = [Vec2(0, y), Vec2(len(lines[y]), y)]
            self.cursor.position = self.selection[1]
            self.draw_selection()
            return

        if key in self.shortcuts['copy']:# TODO
            selectedText = self.get_selected()
            if selectedText:
                pyperclip.copy(selectedText)
                # print('-----copy:', selectedText)

        if key in self.shortcuts['paste']:# TODO
            if self.selection:
                self.delete_selected()
            self.add_text(pyperclip.paste())

        if key in self.shortcuts['cut']:# TODO
            selectedText = self.get_selected()
            if selectedText:
                pyperclip.copy(selectedText)
                self.delete_selected()
                # print('-----cut:', selectedText)

        if key in self.shortcuts['select_all']:
            self.select_all()
            return

        if self.register_mouse_input:
            if key == 'left mouse down' and mouse.hovered_entity == self.bg and mouse.point is not None:
                cursor.position = self.get_mouse_position()
                self.selection = [self.cursor.position, self.cursor.position]
                self.draw_selection()

            if key == 'left mouse up':
                    cursor.position = self.get_mouse_position()

                    if self.selection:
                        self.selection[1] = self.cursor.position



    def move_to_start_of_word(self):
        cursor = self.cursor
        x, y = int(cursor.x), int(cursor.y)
        lines = self.text.split('\n')
        l = lines[y]
        delimiters = self.delimiters

        if x == 0 and y == 0:
            pass

        elif x == 0 and y > 0:                        # move to end of line above
            cursor.y -= 1
            cursor.x = len(lines[y-1])

        elif l[x-1] not in delimiters:              # move left to closest delimiter
            for x in range(x-1, 0, -1):
                if l[x] in delimiters:
                    cursor.x = x+1
                    return
            cursor.x = 0
        else:                                       # move left to closest word
            for x in range(x-1, 0, -1):
                if l[x] not in delimiters:
                    cursor.x = x+1
                    return
            cursor.x = 0

    def move_to_end_of_word(self):
        cursor = self.cursor
        x, y = int(cursor.x), int(cursor.y)
        lines = self.text.split('\n')
        l = lines[y]
        delimiters = self.delimiters

        if x == len(l):
            if y < len(lines)-1:        # end of line, move to beginning of next
                cursor.y += 1
                cursor.x = 0
            return

        elif l[x] not in delimiters:                # move right to closest delimiter
            for x in range(x, len(l)):
                if l[x] in delimiters:
                    cursor.x = x
                    return
            cursor.x = len(l)
        else:                                       # move right to closest word
            for x in range(x, len(l)):
                if l[x] not in delimiters:
                    cursor.x = x
                    return
            cursor.x = len(l)

    def scroll_to_bottom(self, blank_lines_at_bottom=0):
        # self.scroll = min(len(self.text.split('\n')), self.max_lines)
        self.set_scroll(len(self.text.split('\n'))-self.max_lines+blank_lines_at_bottom)
        # print('scrolled to bottom', min(len(self.text.split('\n')), self.max_lines))


    def text_input(self, key):
        cursor, add_text = self.cursor, self.add_text

        if not self.active:
            return

        if held_keys['shift'] and held_keys['alt']:
            return

        if self.selection:  # delete selected text
            self.delete_selected()

        add_text(key)

        if key == '(':
            add_text(')')
            cursor.x -= 1
        elif key == '[':
            add_text(']')
            cursor.x -= 1
        elif key == '{':
            add_text('}')
            cursor.x -= 1

        self.render()


    def render(self):
        lines = self.text.split('\n')
        if self.max_lines < 9999:
            text = '\n'*self.scroll + '\n'.join(lines[self.scroll : self.max_lines + self.scroll])
        else:
            text = self.text

        if not hasattr(self.text_entity, 'raw_text') or self._prev_text != text or self.scroll != self._prev_scroll:
            # print(self.scroll)

            if self.replacements:
                _lines = text.split('\n')
                for i in range(len(_lines)):
                    if _lines[i].lstrip().startswith('#'):
                        _lines[i] = f'☾gray☽{lines[i]}☾default☽'
                    else:
                        _lines[i] = multireplace(_lines[i], self.replacements)
                self.text_entity.text = '\n'.join(_lines)

                # self.text_entity.text = multireplace(text, self.replacements)
            else:
                self.text_entity.text = text

            self.line_numbers.text = ('     \n'*self.scroll) + '\n'.join([str(e + self.scroll).rjust(3, ' ') for e in range(min(len(lines), self.max_lines))])

            self._prev_text = text
            self._prev_scroll = self.scroll
            self.scroll_parent.y = (self.scroll * Text.size * self.line_height)
            self.cursor.visible = self.cursor.y >= self.scroll and self.cursor.y < self.scroll + self.max_lines
            self.draw_selection()

        if self.on_value_changed:
            self.on_value_changed()



    def update(self):
        if self.active and self.register_mouse_input and mouse.left and mouse.moving:
            self.cursor.position = self.get_mouse_position()
            if self.selection:
                self.selection[1] = self.cursor.position

            self.draw_selection()


    def select_all(self):
        lines = self.text.split('\n')
        if lines:
            self.selection = [Vec2(0,0), Vec2(len(lines[-1]), len(lines) - 1)]

        self.draw_selection()


    def draw_selection(self):
        # print(self.selection)
        [destroy(c) for c in self.selection_parent.children]

        if not self.selection or self.selection[0] == self.selection[1]:
            return

        sel = self._ordered_selection()
        # print('---', self.selection, self.scroll)

        start_y = int(sel[0].y)
        end_y = int(sel[1].y)
        lines = self.text.split('\n')
        if start_y == end_y:
            e = Entity(parent=self.selection_parent, model='quad', origin=(-.5,-.5), color=self.highlight_color, double_sided=True, ignore=True, y=start_y)
            e.x = sel[0].x
            e.scale_x = sel[1].x - sel[0].x
        else:
            # first line
            if start_y >= self.scroll and start_y < self.scroll+self.max_lines:
                e = Entity(parent=self.selection_parent, model='quad', origin=(-.5, -.5),
                color=self.highlight_color, double_sided=True, position=(sel[0].x,start_y), ignore=True)
                e.scale_x = len(lines[start_y]) - sel[0].x

            # middle lines
            for y in range(max(start_y+1, self.scroll), min(end_y, self.scroll+self.max_lines)):
                e = Entity(parent=self.selection_parent, model='quad', origin=(-.5, -.5),
                    color=self.highlight_color, double_sided=True, position=(0,y), ignore=True, scale_x=len(lines[y]))

            # last line
            if end_y >= self.scroll and end_y < self.scroll+self.max_lines:
                e = Entity(parent=self.selection_parent, model='quad', origin=(-.5, -.5),
                color=self.highlight_color, double_sided=True, position=(0,end_y), ignore=True)
                e.scale_x = sel[1].x

        for e in self.selection_parent.children:
            e.enabled = e.y >= self.scroll and e.y < self.scroll+self.max_lines


if __name__ == '__main__':
    from ursina import Ursina, window, Button
    app = Ursina(vsync=60)

    # camera.orthographic = True
    # camera.fov = 1
    # window.size = window.fullscreen_size
    # window.x = 200

    window.color = color.hsv(0, 0, .1)
    Button.default_color = color._20
    window.color = color._25

    # Text.size = 1/window.fullscreen_size[1]*16
    # Text.default_font = 'consola.ttf'
    # Text.default_resolution = 16*2
    # TreeView()
    te = TextField(max_lines=30, scale=1, register_mouse_input = True, text='1234')
    #te = TextField(max_lines=300, scale=1, register_mouse_input = True, scroll_size = (50,3))
    # te.line_numbers.enabled = True
    # for name in color.color_names:
    #     if name == 'black':
    #         continue
    #     te.replacements[f' {name}'] = f'☾{name}☽ {name}☾default☽'
    # te.replacements = {
    #     'class ':    '☾orange☽class ☾default☽',
    #     'def ':      '☾azure☽def ☾default☽',
    #     '__init__':  '☾cyan☽__init__☾default☽',
    #     'Entity':    '☾lime☽Entity☾default☽',
    #     'self.':     '☾orange☽self☾default☽.',
    #     '(self)':     '(☾orange☽self☾default☽)',
    #     'self,':     '☾orange☽self☾default☽,',
    #     '    ':      '☾dark_gray☽----☾default☽',
    #     }
    #
    from textwrap import dedent
    te.text = dedent('''
        Lorem ipsum dolor sit amet, consectetur adipiscing elit.
        Aliquam sapien tellus, venenatis sit amet ante et, malesuada porta risus.
        Etiam et mi luctus, viverra urna at, maximus eros. Sed dictum faucibus purus,
        nec rutrum ipsum condimentum in. Mauris iaculis arcu nec justo rutrum euismod.
        Suspendisse dolor tortor, congue id erat sit amet, sollicitudin facilisis velit.
        Aliquam sapien tellus, venenatis sit amet ante et, malesuada porta risus.
        Etiam et mi luctus, viverra urna at, maximus eros. Sed dictum faucibus purus,
        nec rutrum ipsum condimentum in. Mauris iaculis arcu nec justo rutrum euismod.
        Suspendisse dolor tortor, congue id erat sit amet, sollicitudin facilisis velit.
        '''*30
        )[1:]
    te.render()

    def input(key):
        if key == '3':
            te.input('scroll down')


    app.run()



from ursina import *


app = Ursina()

e = Entity(model='quad', texture='heightmap')
tex = load_texture('brick')
print('----------------', tex)

verts = list()
tris = list()
colors = list()
uvs = list()

skip = 10
i = 0
for y in range(0, e.texture.height-skip, skip):
    for x in range(0, e.texture.width-skip, skip):
        c = e.texture.get_pixel(x,y)
        # nc = e.texture.get_pixel(x+1, y)
        # verts.append((x/skip, color.brightness(e.texture.get_pixel(x,y+1)), (y/skip +1)))
        # verts.append((x/skip, color.brightness(c), y/skip))
        #
        # colors.append(e.texture.get_pixel(x,y+1))
        # colors.append(c)

        verts.append((x/skip, color.brightness(c), y/skip))
        verts.append(((x/skip) +1, color.brightness(e.texture.get_pixel(x+skip,y)), y/skip))
        verts.append(((x/skip) +1, color.brightness(e.texture.get_pixel(x+skip,y+skip)), (y/skip) +1))
        verts.append((x/skip, color.brightness(e.texture.get_pixel(x,y+skip)), (y/skip) +1))

        # verts.append((x/skip, color.brightness(c), y/skip))
        # verts.append(((x/skip) +1, color.brightness(e.texture.get_pixel(x+skip,y+skip)), (y/skip) +1))
        # verts.append((x/skip, color.brightness(e.texture.get_pixel(x,y+skip)), (y/skip) +1))

        for t in (0,1,2, 0,2,3):
            tris.append(i+t)

        uvs.append((x/e.texture.width, y/e.texture.height))
        uvs.append(((x+1) /e.texture.width, y/e.texture.height))
        uvs.append(((x+1) /e.texture.width, (y+1) /e.texture.height))
        uvs.append((x /e.texture.width, (y+1) /e.texture.height))
        #
        # uvs.append((x /e.texture.width, y /e.texture.height))
        # uvs.append(((x+1) /e.texture.width, (y+1) /e.texture.height))
        # uvs.append((x /e.texture.width, (y+1) /e.texture.height))

        # uvs.append((0,0))
        # uvs.append((1,0))
        # uvs.append((1,1))
        # uvs.append((0,1))
        #
        # uvs.append((0,0))
        # uvs.append((1,1))
        # uvs.append((0,1))

        # for j in range(4):
        # colors.append(c)
        # colors.append(e.texture.get_pixel(x+skip,y))
        # colors.append(e.texture.get_pixel(x+skip,y+skip))
        # colors.append(e.texture.get_pixel(x,y+skip))

        i += 4
        # print((x /e.texture.width, (y) /e.texture.height))

# e.model = Mesh(vertices=verts, colors=colors, mode='tristrip')
e.model = Mesh(vertices=verts, triangles=tris, colors=colors, uvs=uvs, mode='triangle')
e.texture = 'heightmap'
# e.double_sided = False
e.scale_y *= 10
e.scale *= .08
Sky()
EditorCamera()

class Water(Entity):
    def __init__(self):
        super().__init__()
        self.model = 'plane'
        self.scale *= 100
        # self.color=color.red
        self.texture ='default_sky'
    def update(self):
        self.y += held_keys['up arrow'] * .005
        self.y -= held_keys['down arrow'] * .005
Water()

# window.size = (450, 450/window.aspect_ratio)
# VideoRecorder()
app.run()





def volume(e):
    return e[0] * e[1]


def create_shapes(size_x, size_y):
    sizes = [(size_x, size_y) for i in range(size_x*size_y)]

    i = 0
    for x in range(size_x):
        for y in range(size_y):
            sizes[i] = (x+1,y+1)
            i += 1
    sizes.sort(key=volume, reverse=True)
    return sizes


def grid_to_rects(grid, solid_value=1):
    '''
    Converts a 2d grid into rects by finding the biggest shapes first.
    Parameters: grid(2D array), grid size x(int), grid_size_y(int)
    Returns: a list of (position, size)
    '''
    grid_size_x = len(grid)
    grid_size_y = len(grid[0])

    shapes = create_shapes(grid_size_x, grid_size_y)
    filled = [[0 for y in range(grid_size_y)] for x in range(grid_size_x)]

    # find number of blocks so we can exit early when all have been found
    num_solid_blocks = 0
    for x in range(grid_size_x):
        for y in range(grid_size_y):
            if grid[x][y] == solid_value:
                num_solid_blocks += 1


    rects = []

    for shape in shapes:
        w, h = shape
        # print('checking shape:', w,h,d)
        # check at each position
        for x in range(grid_size_x - w + 1):
            for y in range(grid_size_y - h + 1):
                if filled[x][y] == solid_value:
                    continue

                if shape_fits_in_grid(grid, x,y, filled, grid_size_x, grid_size_y, w, h, solid_value):
                    rects.append(((x,y), shape))

                    for _x in range(w):
                        for _y in range(h):
                            filled[x+_x][y+_y] = 1
                            num_solid_blocks -= 1

                    if num_solid_blocks == 0:
                        return rects

    return rects


def shape_fits_in_grid(grid, start_x, start_y, filled, grid_size_x, grid_size_y, shape_w, shape_h, solid_value):
    for x in range(shape_w):
        for y in range(shape_h):
            if filled[start_x+x][start_y+y]:
                return False
            if not grid[start_x+x][start_y+y] == solid_value:
                return False
    return True


if __name__ == '__main__':
    from time import perf_counter

    def string_to_2d_array(str):
        # print(str)

        lines = str.split('\n')
        lines.reverse()
        w = len(lines[0])
        h = len(lines)
        # print('----', lines)

        grid =  [[0 for y in range(h)] for x in range(w)]
        for y, line in enumerate(lines):
            for x, char in enumerate(line):
                grid[x][y] = int(char)

        # print('----', grid)

        return grid

    from textwrap import dedent
    grid = string_to_2d_array(dedent('''
        0002222222222222
        2222222211111112
        2222222211221112
        2222222111221112
        2222222211221112
        2221112222222112
        2221112222222112
        2221112222222111
        2222222222222111
        ''').strip())



    print(grid)

    t = perf_counter()
    rects = grid_to_rects(grid, solid_value=1)
    print('--------', perf_counter() - t)
    print(rects)


    from ursina import *
    app = Ursina()

    entities = []
    for e in rects:
        pos, scale = e
        entities.append(Entity(model='quad', origin=(-.5,-.5,-.5), position=pos, scale=scale, texture='white_cube'))


    EditorCamera()
    app.run()



from ursina import *; height_shader = Shader(language=Shader.GLSL, vertex = '''#version 140
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelMatrix;
uniform mat4 p3d_ModelViewMatrix;
uniform mat3 p3d_NormalMatrix;
in vec4 p3d_Vertex;
in vec3 p3d_Normal;
in vec2 p3d_MultiTexCoord0;

out vec2 uv;

void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    uv = vec2(.5, p3d_Vertex.y * p3d_MultiTexCoord0.y) + vec2(.5,.5);
}
''',
fragment='''
#version 130
uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;

in vec2 uv;
out vec4 fragColor;

void main() {
    vec3 base = texture2D( p3d_Texture0, uv ).rgb;
    fragColor = vec4( base, 1. ) * p3d_ColorScale;
}

''',
)



if __name__ == '__main__':
    '''
    use matcap textures
    '''
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina()
    window.color=color.black

    shader = height_shader

    a = WhiteCube(shader=shader, texture='matcap_4')
    b = WhiteSphere(shader=shader, rotation_y=180, x=3, texture='matcap_4')
    # AzureSphere(shader=a.shader, y=2)
    GrayPlane(scale=10, y=-2, texture='shore')

    Sky(color=color.light_gray)
    EditorCamera()

    # def update():
    #     b.rotation_z += 1
    #     b.rotation_y += 1
    #     b.rotation_x += 1
    # EditorCamera()

    app.run()



from ursina import *; matcap_shader = Shader(language=Shader.GLSL, fragment='''
#version 130

// HQ2x
// by Lior Halphon
// ported to RetroArch's shader formats by hunterk

/* Based on this (really good) article: http://blog.pkh.me/p/19-butchering-hqx-scaling-filters.html */

/*
MIT License
Copyright (c) 2015-2016 Lior Halphon
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

uniform int FrameDirection;
uniform int FrameCount;
uniform vec2 OutputSize;
uniform vec2 TextureSize;
uniform vec2 InputSize;
// uniform sampler2D Texture;
uniform sampler2D p3d_Texture0;
in vec2 uv;
out vec4 frag_color;
// COMPAT_VARYING vec4 TEX0;

// fragment compatibility #defines
// #define Source Texture
// #define vTexCoord TEX0.xy
vec2 textureDimensions = vec2(16, 16);
// vec2 OutputSize = vec2(512,512);

#define SourceSize vec4(TextureSize, 1.0 / TextureSize) //either TextureSize or InputSize
#define outsize vec4(OutputSize, 1.0 / OutputSize)

/* The colorspace used by the HQnx filters is not really YUV, despite the algorithm description claims it is. It is
   also not normalized. Therefore, we shall call the colorspace used by HQnx "HQ Colorspace" to avoid confusion. */
vec3 rgb_to_hq_colospace(vec4 rgb)
{
    return vec3( 0.250 * rgb.r + 0.250 * rgb.g + 0.250 * rgb.b,
                 0.250 * rgb.r - 0.000 * rgb.g - 0.250 * rgb.b,
                -0.125 * rgb.r + 0.250 * rgb.g - 0.125 * rgb.b);
}

bool is_different(vec4 a, vec4 b)
{
    vec3 diff = abs(rgb_to_hq_colospace(a) - rgb_to_hq_colospace(b));
    return diff.x > 0.188 || diff.y > 0.027 || diff.z > 0.031;
}

#define P(m, r) ((pattern & (m)) == (r))

vec4 interp_2px(vec4 c1, float w1, vec4 c2, float w2)
{
    return (c1 * w1 + c2 * w2) / (w1 + w2);
}

vec4 interp_3px(vec4 c1, float w1, vec4 c2, float w2, vec4 c3, float w3)
{
    return (c1 * w1 + c2 * w2 + c3 * w3) / (w1 + w2 + w3);
}

vec4 scale(sampler2D image, vec2 coord)
{
    // o = offset, the width of a pixel
    vec2 o = 1.0 / textureDimensions;
    vec2 texCoord = coord;

    /* We always calculate the top left pixel.  If we need a different pixel, we flip the image */

    // p = the position within a pixel [0...1]
    vec2 p = fract(texCoord * textureDimensions);

    if (p.x > 0.5) o.x = -o.x;
    if (p.y > 0.5) o.y = -o.y;

    vec4 w0 = texture(image, texCoord + vec2( -o.x, -o.y));
    vec4 w1 = texture(image, texCoord + vec2(    0, -o.y));
    vec4 w2 = texture(image, texCoord + vec2(  o.x, -o.y));
    vec4 w3 = texture(image, texCoord + vec2( -o.x,    0));
    vec4 w4 = texture(image, texCoord + vec2(    0,    0));
    vec4 w5 = texture(image, texCoord + vec2(  o.x,    0));
    vec4 w6 = texture(image, texCoord + vec2( -o.x,  o.y));
    vec4 w7 = texture(image, texCoord + vec2(    0,  o.y));
    vec4 w8 = texture(image, texCoord + vec2(  o.x,  o.y));

    int pattern = 0;
    if (is_different(w0, w4)) pattern |= 1;
    if (is_different(w1, w4)) pattern |= 2;
    if (is_different(w2, w4)) pattern |= 4;
    if (is_different(w3, w4)) pattern |= 8;
    if (is_different(w5, w4)) pattern |= 16;
    if (is_different(w6, w4)) pattern |= 32;
    if (is_different(w7, w4)) pattern |= 64;
    if (is_different(w8, w4)) pattern |= 128;

    if ((P(0xbf,0x37) || P(0xdb,0x13)) && is_different(w1, w5))
        return interp_2px(w4, 3.0, w3, 1.0);
    if ((P(0xdb,0x49) || P(0xef,0x6d)) && is_different(w7, w3))
        return interp_2px(w4, 3.0, w1, 1.0);
    if ((P(0x0b,0x0b) || P(0xfe,0x4a) || P(0xfe,0x1a)) && is_different(w3, w1))
        return w4;
    if ((P(0x6f,0x2a) || P(0x5b,0x0a) || P(0xbf,0x3a) || P(0xdf,0x5a) ||
         P(0x9f,0x8a) || P(0xcf,0x8a) || P(0xef,0x4e) || P(0x3f,0x0e) ||
         P(0xfb,0x5a) || P(0xbb,0x8a) || P(0x7f,0x5a) || P(0xaf,0x8a) ||
         P(0xeb,0x8a)) && is_different(w3, w1))
        return interp_2px(w4, 3.0, w0, 1.0);
    if (P(0x0b,0x08))
        return interp_3px(w4, 2.0, w0, 1.0, w1, 1.0);
    if (P(0x0b,0x02))
        return interp_3px(w4, 2.0, w0, 1.0, w3, 1.0);
    if (P(0x2f,0x2f))
        return interp_3px(w4, 1.04, w3, 1.0, w1, 1.0);
    if (P(0xbf,0x37) || P(0xdb,0x13))
        return interp_3px(w4, 5.0, w1, 2.0, w3, 1.0);
    if (P(0xdb,0x49) || P(0xef,0x6d))
        return interp_3px(w4, 5.0, w3, 2.0, w1, 1.0);
    if (P(0x1b,0x03) || P(0x4f,0x43) || P(0x8b,0x83) || P(0x6b,0x43))
        return interp_2px(w4, 3.0, w3, 1.0);
    if (P(0x4b,0x09) || P(0x8b,0x89) || P(0x1f,0x19) || P(0x3b,0x19))
        return interp_2px(w4, 3.0, w1, 1.0);
    if (P(0x7e,0x2a) || P(0xef,0xab) || P(0xbf,0x8f) || P(0x7e,0x0e))
        return interp_3px(w4, 2.0, w3, 3.0, w1, 3.0);
    if (P(0xfb,0x6a) || P(0x6f,0x6e) || P(0x3f,0x3e) || P(0xfb,0xfa) ||
        P(0xdf,0xde) || P(0xdf,0x1e))
        return interp_2px(w4, 3.0, w0, 1.0);
    if (P(0x0a,0x00) || P(0x4f,0x4b) || P(0x9f,0x1b) || P(0x2f,0x0b) ||
        P(0xbe,0x0a) || P(0xee,0x0a) || P(0x7e,0x0a) || P(0xeb,0x4b) ||
        P(0x3b,0x1b))
        return interp_3px(w4, 2.0, w3, 1.0, w1, 1.0);

    return interp_3px(w4, 6.0, w3, 1.0, w1, 1.0);
}

void main()
{
	frag_color = scale(p3d_Texture0, uv);
}

''',
)



if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina()
    window.color=color.black

    shader = matcap_shader
    Texture.default_filtering = 'bilinear'
    a = Entity(model='quad', shader=shader, texture='brick')
    EditorCamera()

    app.run()



import time
import platform

from direct.showbase.ShowBase import ShowBase
from direct.task.Task import Task
from panda3d.core import ConfigVariableBool

from ursina.window import instance as window
from ursina import application
from ursina import input_handler
from ursina.scene import instance as scene
from ursina.camera import instance as camera
from ursina.mouse import instance as mouse
from ursina import entity
from ursina import shader


import __main__
time.dt = 0
time.dt_unscaled = 0
keyboard_keys = '1234567890qwertyuiopasdfghjklzxcvbnm'


from ursina.scripts.singleton_decorator import singleton
@singleton
class Ursina(ShowBase):
    def __init__(self, title='ursina', icon='textures/ursina.ico', borderless:bool=None, fullscreen:bool=None, size=None, forced_aspect_ratio=None, position=None, vsync=True, editor_ui_enabled:bool=None, window_type='onscreen', development_mode:bool=None, render_mode=None, show_ursina_splash=False, use_ingame_console=False, **kwargs):
        """The main class of Ursina. This class is a singleton, so you can only have one instance of it.

        Keyword Args (optional):
            title (str): The title of the window.\n
            fullscreen (bool): Whether the window should be fullscreen or not.\n
            size (tuple(int, int)): The size of the window.\n
            forced_aspect_ratio (bool): Whether the window should have a forced aspect ratio or not.\n
            position (tuple(int, int)): The position of the window. Defaults to center of main monitor.\n
            vsync (bool): Whether the window should have vsync enabled or not.\n
            borderless (bool): Whether the window should be borderless or not.\n
            show_ursina_splash (bool): Whether the Ursina splash should be shown or not.\n
            render_mode (str): The render mode of the window.\n
            development_mode (bool): Whether the development mode should be enabled or not.\n
            editor_ui_enabled (bool): Whether the editor UI should be enabled or not.\n
            window_type (str): The type of the window. Can be 'onscreen', 'offscreen' or 'none'.\n
        """
        entity._warn_if_ursina_not_instantiated = False
        application.window_type = window_type
        application.base = self
        if development_mode is not None:
            application.development_mode = development_mode

        application.show_ursina_splash = show_ursina_splash

        try:
            import gltf
            gltf.patch_loader(self.loader)
        except:
            pass

        if 'gltf_no_srgb' in kwargs:
            application.gltf_no_srgb = kwargs['gltf_no_srgb']

        if editor_ui_enabled is None:
            editor_ui_enabled = bool(application.development_mode)

        if fullscreen is None and not application.development_mode:
            fullscreen = True
            if borderless is None:
                borderless = True

        window._ready(title=title, icon=icon,
            borderless=borderless, fullscreen=fullscreen, size=size, forced_aspect_ratio=forced_aspect_ratio, position=position, vsync=vsync, window_type=window_type,
            editor_ui_enabled=editor_ui_enabled, render_mode=render_mode)

        super().__init__(windowType=application.window_type)
        window.apply_settings()
        # camera
        if application.window_type != 'none':
            camera._cam = self.camera
            camera._cam.reparent_to(camera)
            camera.render = self.render
            camera.position = (0, 0, -20)
            scene.camera = camera
            camera._set_up()

        # input
        if application.window_type == 'onscreen':
            self.buttonThrowers[0].node().setButtonDownEvent('buttonDown')
            self.buttonThrowers[0].node().setButtonUpEvent('buttonUp')
            self.buttonThrowers[0].node().setButtonRepeatEvent('buttonHold')
            self.buttonThrowers[0].node().setKeystrokeEvent('keystroke')
        self._input_name_changes = {
            'mouse1' : 'left mouse down', 'mouse1 up' : 'left mouse up', 'mouse2' : 'middle mouse down', 'mouse2 up' : 'middle mouse up', 'mouse3' : 'right mouse down', 'mouse3 up' : 'right mouse up',
            'wheel_up' : 'scroll up', 'wheel_down' : 'scroll down',
            'arrow_left' : 'left arrow', 'arrow_left up' : 'left arrow up', 'arrow_up' : 'up arrow', 'arrow_up up' : 'up arrow up', 'arrow_down' : 'down arrow', 'arrow_down up' : 'down arrow up', 'arrow_right' : 'right arrow', 'arrow_right up' : 'right arrow up',
            'lcontrol' : 'left control', 'rcontrol' : 'right control', 'lshift' : 'left shift', 'rshift' : 'right shift', 'lalt' : 'left alt', 'ralt' : 'right alt',
            'lcontrol up' : 'left control up', 'rcontrol up' : 'right control up', 'lshift up' : 'left shift up', 'rshift up' : 'right shift up', 'lalt up' : 'left alt up', 'ralt up' : 'right alt up',
            'control-mouse1' : 'left mouse down', 'control-mouse2' : 'middle mouse down', 'control-mouse3' : 'right mouse down',
            'shift-mouse1' : 'left mouse down', 'shift-mouse2' : 'middle mouse down', 'shift-mouse3' : 'right mouse down',
            'alt-mouse1' : 'left mouse down', 'alt-mouse2' : 'middle mouse down', 'alt-mouse3' : 'right mouse down',
            'control-alt-mouse1' : 'left mouse down', 'control-alt-mouse2' : 'middle mouse down', 'control-alt-mouse3' : 'right mouse down',
            'shift-control-alt-mouse1' : 'left mouse down', 'shift-control-alt-mouse2' : 'middle mouse down', 'shift-control-alt-mouse3' : 'right mouse down',
            'page_down' : 'page down', 'page_down up' : 'page down up', 'page_up' : 'page up', 'page_up up' : 'page up up',
            }

        for e in keyboard_keys:
            self.accept(f'raw-{e}', self.input, [e, True])
            self.accept(f'raw-{e}-up', self.input_up, [e, True])
            self.accept(f'raw-{e}-repeat', self.input_hold, [e, True])

        self.accept('buttonDown', self.input)
        self.accept('buttonUp', self.input_up)
        self.accept('buttonHold', self.input_hold)
        self.accept('keystroke', self.text_input)
        ConfigVariableBool('paste-emit-keystrokes', False)

        self.disableMouse()
        mouse._mouse_watcher = self.mouseWatcherNode
        mouse.enabled = True
        self.mouse = mouse

        scene._set_up()

        # try to load settings that need to be applied before entity creation
        application.load_settings()

        from ursina.prefabs.hot_reloader import HotReloader
        application.hot_reloader = HotReloader(__main__.__file__ if hasattr(__main__, '__file__') else 'None', enabled=application.development_mode)  # make sure it's running from a file and not an interactive session.

        try:
            from ursina import gamepad
        except Exception as e:
            print(e)

        if application.window_type != 'none':
            window.make_editor_gui()
            if use_ingame_console:
                import builtins
                from ursina import Entity, TextField, color
                window.console = Entity(parent=window.editor_ui, position=window.top_left, z=-999, eternal=True)
                window.console.text_field = TextField(parent=window.console, scale=.75, max_lines=20, position=(0,0), register_mouse_input=True, text_input=None, eternal=True)
                window.console.text_field.bg.color = color.black66
                window.console.text_field.bg.scale_x = 1.5
                def _custom_print(*args, **kwargs):  # makes print() poutput to the in-game console instead of the terminal if use_ingame_console is True
                    content = ' '.join([str(e) for e in args])
                    if 'error' in content:
                        content = f'{content}'
                    window.console.text_field.cursor.y = window.console.text_field.text.count('\n')
                    window.console.text_field.cursor.x = 999
                    window.console.text_field.add_text(f'\n{content}')
                    window.console.text_field.scroll_to_bottom(1)
                    window.console.text_field.render()
                builtins.print = _custom_print

                def _console_text_input(key):
                    if key == '|':
                        window.console.text_field.enabled = not window.console.text_field.enabled
                window.console.text_input = _console_text_input

        print('package_folder:', application.package_folder)
        print('asset_folder:', application.asset_folder)

        entity._Ursina_instance = self
        self._update_task = self.taskMgr.add(self._update, "update")

        # since opening a window is non-blocking, we need a way to call functions exactly when the window is ready, for example to make the splash screen animation not finish before you get to see it, or the camera to move before you see anything.
        def _wait_for_window_open(task):
            if self.win and self.win.isValid():
                if hasattr(__main__, 'on_window_ready') and __main__.on_window_ready:
                    __main__.on_window_ready()

                for e in scene.entities:
                    if hasattr(e, 'on_window_ready') and callable(e.on_window_ready):
                        e.on_window_ready()
                return Task.done
            return Task.cont
        self.taskMgr.add(_wait_for_window_open, 'wait_for_window')


    def _update(self, task):
        """Internal task that runs every frame. Updates time, mouse, sequences and entities."""
        if application.calculate_dt:
            time.dt_unscaled = globalClock.getDt()
            time.dt = time.dt_unscaled * application.time_scale          # time between frames
        mouse.update()

        if hasattr(__main__, 'update') and __main__.update and not application.paused:
            __main__.update()

        for seq in application.sequences:
            seq.update()

        if scene._entities_marked_for_removal:
            scene.entities = [e for e in scene.entities if e not in scene._entities_marked_for_removal]
            scene._entities_marked_for_removal.clear()

        for e in scene.entities:
            if not e.enabled or e.ignore:
                continue
            if e in scene._entities_marked_for_removal:
                continue
            if application.paused and e.ignore_paused is False:
                continue
            if e.has_disabled_ancestor():
                continue

            if hasattr(e, 'update') and callable(e.update):
                e.update()

            if hasattr(e, 'scripts'):
                for script in e.scripts:
                    if script.enabled and hasattr(script, 'update') and callable(script.update):
                        script.update()

            if e.shader and hasattr(e.shader, "continuous_input"):
                for key, value in e.shader.continuous_input.items():
                    e.set_shader_input(key, value())

        return Task.cont


    def input_up(self, key, is_raw=False): # internal method for key release
        if not is_raw and key in keyboard_keys:
            return

        if key in ('wheel_up', 'wheel_down'):
            return

        key += ' up'
        self.input(key)


    def input_hold(self, key, is_raw=False):   # internal method for handling repeating input that occurs when you hold the key
        key = key.replace('control-', '')
        key = key.replace('shift-', '')
        key = key.replace('alt-', '')

        if key in self._input_name_changes:
            key = self._input_name_changes[key]

        key += ' hold'
        self.input(key)


    def input(self, key, is_raw=False): # internal method for handling input
        """Built-in input handler. Propagates the input to all entities and the input function of the main script. Main use case for this it to simulate input though code, like: app.input('a').

        Args:
            key (Any): The input key.
            is_raw (bool, optional): Whether or not the input should be treated as "raw". Defaults to False.
        """
        if not is_raw and key in keyboard_keys:
            return

        if 'mouse' not in key:
            for prefix in ('control-', 'shift-', 'alt-'):
                if key.startswith(prefix):
                    key = key.replace('control-', '')
                    key = key.replace('shift-', '')
                    key = key.replace('alt-', '')
                    if key in keyboard_keys:
                        return

        if key in self._input_name_changes:
            key = self._input_name_changes[key]

        # since we can rebind one key to multiple, get a list of keys
        bound_keys = input_handler.rebinds.get(key, (key, ))

        for key in bound_keys:
            input_handler.input(key)

        if not application.paused:
            if hasattr(__main__, 'input'):
                for key in bound_keys:
                    __main__.input(key)

        break_outer = False


        for e in scene.entities:
            if e.enabled is False or e.ignore or e.ignore_input:
                continue
            if application.paused and e.ignore_paused is False:
                continue
            if e.has_disabled_ancestor():
                continue

            if break_outer:
                break

            if hasattr(e, 'input') and callable(e.input):
                for key in bound_keys:
                    if break_outer:
                        break
                    if e.input(key):    # if the input function returns True, eat the input
                        break_outer = True
                        break

            if hasattr(e, 'scripts'):
                if break_outer:
                    break

                for script in e.scripts:
                    if break_outer:
                        break

                    if script.enabled and hasattr(script, 'input') and callable(script.input):
                        for key in bound_keys:
                            if script.input(key): # if the input function returns True, eat the input
                                break_outer = True
                                break

        for key in bound_keys:
            mouse.input(key)


    def text_input(self, key):  # internal method for handling text input
        key_code = ord(key)
        if key_code < 32 or (key_code >= 127 and key_code <= 160):
            return

        if key == ' ' and input_handler.held_keys['control']:
            return

        if not application.paused:
            if hasattr(__main__, 'text_input'):
                __main__.text_input(key)

        for e in scene.entities:
            if e.enabled is False or e.ignore or e.ignore_input:
                continue
            if application.paused and e.ignore_paused is False:
                continue

            if hasattr(e, 'text_input') and callable(e.text_input):
                e.text_input(key)

            if hasattr(e, 'scripts'):
                for script in e.scripts:
                    if script.enabled and hasattr(script, 'text_input') and callable(script.text_input):
                        script.text_input(key)

    def step(self): # use this control the update loop yourself. call app.step() in a while loop for example, instead of app.run()
        self.taskMgr.step()


    def run(self, info=True):
        if application.show_ursina_splash:
            from ursina.prefabs.splash_screen import UrsinaSplashScreen
            application.ursina_splash = UrsinaSplashScreen()

        application.load_settings()
        if info:
            print('os:', platform.system())
            import sys
            print(f'python version: {sys.version}')
            from importlib.metadata import version
            print(f'ursina version: {version('ursina')}')
            print('development mode:', application.development_mode)
            print('application successfully started')

        super().run()


if __name__ == '__main__':
    from ursina import *
    app = Ursina(
        # development_mode=False,
        # use_ingame_console=True,
        # show_ursina_splash=True,
    )
    def input(key):
        print(key)
    app.run()





if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    from panda3d.core import ColorBlendAttrib
    app = Ursina()

    e = Entity(model='sphere', texture='shore', y=2, alpha=.4)
    e.setAttrib(ColorBlendAttrib.make(ColorBlendAttrib.MAdd))
    ground = GrayPlane(scale=10, y=-2, texture='shore', shader=shader, texture_scale=(10,10))
    EditorCamera()


    app.run()



import sys
sys.path.append("..")
from ursina import *

class NewScene():

    def __init__(self):
        self.is_editor = True
        self.target = None

        self.waiting_for_reply = False
        # self.submit_name = scene.editor.save_scene_button.save_scene_button.ask_for_scene_name_menu.save_button


    def input(self, key):
        if key == 'left mouse down' and self.entity.hovered:
            # self.target.enabled = not self.target.enabled
            if scene.has_changes:
                # ask for save / discard

                self.waiting_for_reply = True

            # ask for scene name

            destroy(scene.entity)
            scene.entity = Entity()
            scene.entity.name = 'untitled_scene'
            scene.editor.hierarchy_panel_header.text = scene.entity.name

            # save_prefab

    def update(self):
        if self.waiting_for_reply:
            pass



from ursina import Shader

camera_empty_shader = Shader(
vertex='''
#version 430

in vec4 p3d_Vertex;
uniform mat4 p3d_ViewMatrixInverse;
in vec2 p3d_MultiTexCoord0;
out vec2 uv;

void main() {
    gl_Position = p3d_ViewMatrixInverse  * p3d_Vertex;
    uv = p3d_MultiTexCoord0;
}
''',

fragment='''
#version 430

uniform sampler2D tex;
in vec2 uv;
out vec4 color;

void main() {
    vec3 rgb = texture(tex, uv).rgb;
    color = vec4(rgb, 1.0);
}

''')



if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere', color=color.yellow)
    e = Entity(model='cube', y=-1)
    camera.shader = camera_empty_shader
    camera.set_shader_input('contrast', 1)

    #slider = ThinSlider(max=1, dynamic=True, position=(-.25, -.45))
    #def set_blur():
    #    camera.set_shader_input("contrast", slider.value)
    #slider.on_value_changed = set_blur

    EditorCamera()
    Sky()
    app.run()



from math import sqrt, sin, acos, pi, cos, floor
from math import hypot
from panda3d.core import Vec4, LVector3f
from ursina.vec2 import Vec2
from ursina.vec3 import Vec3
from ursina import color
from ursina.color import Color
_sum = sum


def distance(a, b):
    if isinstance(a, (Color, Vec4)):
        dist = abs(a[0] - b[0])
        dist += abs(a[1] - b[1])
        dist += abs(a[2] - b[2])
        dist += abs(a[3] - b[3])
        return dist

    # if input is Entity, convert to positions
    if hasattr(a, 'world_position'): a = a.world_position
    if hasattr(b, 'world_position'): b = b.world_position

    dist = sqrt((b[0] - a[0])**2 + (b[1] - a[1])**2 + (b[2] - a[2])**2)
    # print('------------DIST:', dist)
    return dist


def distance_2d(a, b):
    if hasattr(a, 'position'): a = a.position
    if hasattr(b, 'position'): b = b.position

    return sqrt((b[0] - a[0])**2 + (b[1] - a[1])**2)


def distance_xz(a, b):
    if hasattr(a, 'position'): a = a.position
    if hasattr(b, 'position'): b = b.position

    return sqrt((b[0] - a[0])**2 + (b[2] - a[2])**2)


def lerp(a, b, t):
    if isinstance(a, (int, float, complex)):
        return a + (b - a) * t

    elif isinstance(a, Color) and isinstance(b, Color):
        col = [lerp(e[0], e[1], t) for e in zip(a,b)]
        return Color(col[0], col[1], col[2], col[3])

    elif isinstance(a, (tuple, list, Vec2, Vec3, Vec4, LVector3f)) and isinstance(b, (tuple, list, Vec2, Vec3, Vec4, LVector3f)):
        lerped = []
        for i in range(min(len(a), len(b))):
            lerped.append(lerp(a[i], b[i], t))

        if isinstance(a, (tuple, list)):
            return type(a)(lerped)
        else:
            return type(a)(*lerped)
    else:
        raise TypeError(f'''can't lerp types {type(a)} and {type(b)}''')

def inverselerp(a, b, value):   # get *where* between a and b, value is (0.0 - 1.0)
    if a == b:
        return .5
    return (value - a) / (b - a)

if __name__ == '__main__':
    from ursina.ursinastuff import _test
    _test(inverselerp(0, 100, 50) == .5)
    _test(lerp(0, 100, .5) == 50)


def lerp_exponential_decay(a, b, decay_rate):    # frame-rate independent lerp for use in update. use this instead of lerp(a, b, time.dt) in update.
    return lerp(a, b, 1 - pow(0.01, decay_rate))


def lerp_angle(start_angle, end_angle, t):
    start_angle = start_angle % 360
    end_angle = end_angle % 360
    angle_diff = (end_angle - start_angle + 180) % 360 - 180
    result_angle = start_angle + t * angle_diff
    result_angle = (result_angle + 360) % 360
    return result_angle


def slerp(q1, q2, t):
    costheta = q1.dot(q2)

    # ensure shortest path by flipping q2 if dot product is negative
    if costheta < 0.0:
        q2 = -q2
        costheta = -costheta

    costheta = clamp(costheta, -1.0, 1.0)   # ensure valid range for acos

    theta = acos(costheta)
    if abs(theta) < 0.0001:
        return q2

    sintheta = sqrt(1.0 - costheta * costheta)
    if abs(sintheta) < 0.0001:
        return (q1 + q2) * 0.5

    r1 = sin((1.0 - t) * theta) / sintheta
    r2 = sin(t * theta) / sintheta
    return (q1 * r1) + (q2 * r2)


def slerp_exponential_decay(q1, q2, decay_rate):    # frame-rate independent version of slerp for use in update.
    return slerp(q1, q2, 1 - pow(0.01, decay_rate))


def clamp(value, floor, ceiling):
    return max(min(value, ceiling), floor)


def round_to_closest(value, step=0):
    if not step:
        return value

    step = 1/step
    return round(value * step) / step


def rotate_around_point_2d(point, origin, deg):
    angle_rad = -deg/180 * pi # ursina rotation is positive=clockwise, so do *= -1
    cos_angle = cos(angle_rad)
    sin_angle = sin(angle_rad)
    dx = point[0] - origin[0]
    dy = point[1] - origin[1]

    return (
        origin[0] + (dx*cos_angle - dy*sin_angle),
        origin[1] + (dx*sin_angle + dy*cos_angle)
        )

def world_position_to_screen_position(point): # get screen position(ui space) from world space.
    from ursina import camera, Entity, destroy
    _temp_entity = Entity(position=point, add_to_scene_entities=False)
    result = _temp_entity.screen_position
    destroy(_temp_entity)
    return result


def sum(l):
    try:
        return _sum(l)
    except:
        pass

    total = l[0].__class__()
    for e in l:
        total += e

    return total


def make_gradient(index_value_dict):
    '''
    given a dict of positions and values (usually colors), interpolates the values into a list of with the interpolated values.
    example input: {'0':color.hex('#9d9867'), '38':color.hex('#828131'), '54':color.hex('#5d5b2a'), '255':color.hex('#000000')}
    '''
    min_index = min(int(e) for e in index_value_dict.keys())
    max_index = max(int(e) for e in index_value_dict.keys())
    # default_value = index_value_dict.values()[0]
    # print('-------------', tuple(index_value_dict.values())[0])
    gradient = [None for _ in range(max_index+1-min_index)]

    sorted_dict = [(idx, index_value_dict[str(idx)]) for idx in sorted([int(key) for key in index_value_dict.keys()])]
    # print(sorted_dict)

    for i in range(len(sorted_dict)-1):
        start_index, start_value = sorted_dict[i]
        next_index, next_value = sorted_dict[i+1]
        # print(start_index, '-->', next_index, ':', start_value, '-->', next_value)
        dist = next_index - start_index
        for j in range(dist+1):
            gradient[start_index+j-min_index] = lerp(start_value, next_value, j/dist)

    return gradient

if __name__ == '__main__':
    _test(make_gradient({'0':color.hex('#ff0000ff'), '2':color.hex('#ffffffff')}) == [
        color.hex('#ff0000ff'),
        lerp(color.hex('#ff0000ff'), color.hex('#ffffffff'), .5),
        color.hex('#ffffffff'),
        ])
    _test(make_gradient({'0':color.hex('#ff0000ff'), '4':color.hex('#ffffffff')}) == [
        color.hex('#ff0000ff'),
        lerp(color.hex('#ff0000ff'), color.hex('#ffffffff'), .25),
        lerp(color.hex('#ff0000ff'), color.hex('#ffffffff'), .5),
        lerp(color.hex('#ff0000ff'), color.hex('#ffffffff'), .75),
        color.hex('#ffffffff'),
        ])
    _test(make_gradient({'0':16, '2':0}) == [16, 8, 0])

    _test(make_gradient({'6':0, '8':8}) == [0, 4, 8])


def sample_gradient(list_of_values, t):     # distribute list_of_values equally on a line and get the interpolated value at t (0-1).
    l = len(list_of_values)
    if l == 1:
        return list_of_values[0]

    t *= l-1
    index = floor(t - .001)
    index = clamp(index, 0, l-1)
    relative = t - index
    # print(t, index, relative)

    if index < l-1:
        return lerp(list_of_values[index], list_of_values[index+1], relative)
    else:
        return lerp(list_of_values[index-1], list_of_values[index], relative)



class Bounds:
    __slots__ = ['start', 'end', 'center', 'size']

    def __init__(self, *, start:Vec3=None, end:Vec3=None, center:Vec3=None, size:Vec3=None):
        if start is not None and end is not None:
            self.start = start
            self.end = end
            self.size = end - start
            self.center = start + (self.size * 0.5)

        elif center is not None and size is not None:
            self.center = center
            self.size = size
            half = size * 0.5
            self.start = center - half
            self.end = center + half

        else:
            raise ValueError("Must provide either (start and end) or (center and size)")

    def __repr__(self):
        return f"Bounds(start={self.start}, end={self.end}, center={self.center}, size={self.size})"


if __name__ == '__main__':
    from ursina import *
    from ursinastuff import _test
    app = Ursina()
    e1 = Entity(position = (0,0,0))
    e2 = Entity(position = (0,1,1))
    _test(distance(e1, e2) == 1.4142135623730951)
    _test(distance_2d(Vec2(0,0), Vec2(1,1)) == 1.4142135623730951)

    distance_xz(e1, e2.position)

    between_color = lerp(color.lime, color.magenta, .5)
    print(between_color)
    print(lerp((0,0), (0,1), .5))
    print(lerp(Vec2(0,0), Vec2(0,1), .5))
    print(lerp([0,0], [0,1], .5))

    print(round(Vec3(.38, .1351, 353.26), 2))

    p = (1,0)
    print(p, 'rotated ->', rotate_around_point_2d(p, (0,0), 90))

    app.run()



class every:
    '''@every  decorator for calling a function on an Entity every n seconds.

        example:
        @every(.1)
        def fixed_update():
            print('check collision')

        Using the @every decorator is the same as doing this in __init__() (on Entity):
        self.animations.append(Sequence(Func(self.fixed_update), Wait(.1), loop=True, started=True))
        The Sequence will call the function every .1 second, while adding it to
        self.animations ensures the Sequence gets cleaned up when the Entity gets destroyed.
    '''
    decorated_methods = []  # store decorated methods here

    def __init__(self, interval):
        # print('asdlkjasldkasjdlkj')
        self.interval = interval

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            # print(f"-----------Calling {func.__name__} every {self.interval} seconds")
            return func(*args, **kwargs)

        wrapper._every = self  # add _every attribute to the decorated method
        wrapper._func = func
        every.decorated_methods.append(wrapper)  # store the decorated method
        return wrapper


def get_class_name(func):
    qualname_parts = func.__qualname__.split('.')
    class_name = qualname_parts[-2] if len(qualname_parts) > 1 else None
    return class_name


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    # @every(.1)
    # def test():
    #     print('test')


    class Enemy(Entity):
        @every(.2)
        def attack(self):
            print('attack')

    enemy = Enemy(enabled=1)

    # test that it gets pause when disabling the entity
    def input(key):
        if key == 'space':
            enemy.enabled = not enemy.enabled
            print(enemy.enabled)

    app.run()
